
/* $Id: region.c,v 3.16 1999/11/03 01:33:10 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* System Header Files */

#include <stdio.h>
#include <sys/types.h>

#include <malloc.h>
#include <math.h>

#ifndef PI
#define PI (M_PI)
#endif


/* STV Headers */

#include "stv.h"


/* Communication Constants */

#define APPNAME				"Region"

#define GROUPNAME			"region"

#define STV_MSGTAG			10

#define TRANSFER_MSG		100

#define END_TRANSFER		-1
#define BLIP_TRANSFER		-2
#define BULLET_TRANSFER		-3
#define HELI_TRANSFER		-4
#define SCORE_TRANSFER      -5


/* Simulation Constants */

#define FIELD_HEIGHT		100000
#define FIELD_WIDTH			100000

#define MAX_RADAR_RANGE		10000.0

#define SHIELD_STRENGTH		10

#define HELI_SCORE          10000

#define SPEED_FACTOR		5

#define BULLET_LIFETIME		30

#define BULLET_RADIUS       1

#define BULLET_SCORE		500

#define FIRE_DELAY			10

#define NOISE_LEVEL			0.75

#define EXPLOSION_GROWING      0
#define EXPLOSION_SHRINKING    1

#define HELI_EXPLOSION_STRENGTH      ( 3 * HELI_RADIUS )
#define BULLET_EXPLOSION_STRENGTH    ( HELI_RADIUS / 8 )


/* Macros */

#define NUM_HELI_IDS		100000

#define HELI_ID( _region ) \
( NEXT_HELI_ID = NEXT_HELI_ID >= NUM_HELI_IDS - 1 ? 0 : \
	NEXT_HELI_ID + 1, \
	( ( (_region) + 1 ) * NUM_HELI_IDS ) + NEXT_HELI_ID )

#define DUMP_HELI( _str, _hid, _x, _y, _rr, _h, _hg, _va, _vm, _vg ) \
{ \
	if ( vflag ) \
	{ \
		printf( \
"%s heli %d (%d,%d) R(%.2lf) H(%.2lf/%.2lf) V(%.2lf,%.2lf/%.2lf)\n", \
			_str, _hid, _x, _y, _rr, _h, _hg, _va, _vm, _vg ); \
		fflush( stdout ); \
	} \
}


#define NUM_BULLET_IDS		10000000

#define BULLET_ID( _region ) \
( NEXT_BULLET_ID = NEXT_BULLET_ID >= NUM_BULLET_IDS - 1 ? 0 : \
	NEXT_BULLET_ID + 1, \
	( ( (_region) + 1 ) * NUM_BULLET_IDS ) + NEXT_BULLET_ID )

#define DUMP_BULLET( _str, _bid, _x, _y, _va, _vm ) \
{ \
	if ( vflag ) \
	{ \
		printf( "%s bullet %d (%d,%d) V(%.2lf,%.2lf)\n", \
			_str, _bid, _x, _y, _va, _vm ); \
		fflush( stdout ); \
	} \
}


#define NUM_EXPLOSION_IDS		10000000

#define EXPLOSION_ID( _region ) \
( NEXT_EXPLOSION_ID = NEXT_EXPLOSION_ID >= NUM_EXPLOSION_IDS - 1 ? 0 : \
	NEXT_EXPLOSION_ID + 1, \
	( ( (_region) + 1 ) * NUM_EXPLOSION_IDS ) + NEXT_EXPLOSION_ID )

#define DUMP_EXPLOSION( _str, _eid, _x, _y, _angle, _strength ) \
{ \
	if ( vflag ) \
	{ \
		printf( "%s explosion %d = %d @ (%d,%d) A(%.2lf)\n", \
			_str, _eid, _strength, _x, _y, _angle ); \
		fflush( stdout ); \
	} \
}


#define EQUIV_DELTA			0.000001

#define EQUIV( _d1, _d2 ) \
( ( (_d1) != 0.0 ) ? \
		( ( fabs( (_d1) - (_d2) ) / fabs( _d1 ) ) < EQUIV_DELTA ) \
		: ( fabs( (_d1) - (_d2) ) < EQUIV_DELTA ) )


#define IN_RANGE( _x, _x0, _a ) \
( ( (_a) > 0 ) ? \
	( (_x0) <= (_x) && (_x) <= ( (_x0) + (_a) ) ) \
	: ( (_x0) >= (_x) && (_x) >= ( (_x0) + (_a) ) ) )


/*
#define SUPER_DELTA( _delta, _heli ) \
( (_delta) \
	* ( 1.0 \
		+ ( ( ((double) MAX_RADAR_RANGE) - (_heli)->radar_range ) \
			/ ((double) MAX_RADAR_RANGE) ) ) )
*/

#define SUPER_DELTA( _delta, _heli )    (_delta)


/* Data Structures */

struct region_struct
{
	int tid;
	int xmin, xmax;
	int ymin, ymax;
	int nneigh;
	/* REGION */ struct region_struct *neigh;
};

typedef struct region_struct *REGION;


struct heli_struct
{
	int hid;
	int loc_x, loc_y;
	int fire_count;
	int killed;
	int steered;
	int shield_str;
	int shielded;
	int score;
	int bounce;
	double radar_range;
	double head_goal;
	double heading;
	double vel_angle;
	double vel_magn;
	double vel_goal;
	double wimp_factor;
	double last_bdist;
	/* HELI */ struct heli_struct *next;
};

typedef struct heli_struct *HELI;


struct bullet_struct
{
	int bid;
	int age;
	int loc_x, loc_y;
	double vel_angle;
	double vel_magn;
	int parent;
	/* BULLET */ struct bullet_struct *next;
};

typedef struct bullet_struct *BULLET;


struct explosion_struct
{
	int eid;
	int state;
	int loc_x, loc_y;
	double angle;
	int strength;
	int radius;
	/* EXPLOSION */ struct explosion_struct *next;
};

typedef struct explosion_struct *EXPLOSION;


struct blip_struct
{
	int loc_x, loc_y;
	double radius;
	/* BLIP */ struct blip_struct *next;
};

typedef struct blip_struct *BLIP;


/* Routines */

void	get_helicopter __STV_Prototype__(( int, STV_REGION,
			STV_PARTICLE_ID *, int * ));

void	get_helicopter_pfield __STV_Prototype__(( STV_PARTICLE_ID,
			STV_VALUE, STV_VALUE * ));


void	make_helicopter __STV_Prototype__(( int *, STV_PARTICLE_ID * ));

void	set_helicopter_pfield __STV_Prototype__(( STV_PARTICLE_ID,
			STV_VALUE, STV_VALUE ));


#define PF_HELI_HID				 0
#define PF_HELI_FIRE_COUNT		 1
#define PF_HELI_STEERED			 2
#define PF_HELI_SHIELD_STR		 3
#define PF_HELI_SHIELDED		 4
#define PF_HELI_SCORE			 5
#define PF_HELI_RADAR_RANGE		 6
#define PF_HELI_HEAD_GOAL		 7
#define PF_HELI_HEADING			 8
#define PF_HELI_VEL_ANGLE		 9
#define PF_HELI_VEL_MAGN		10
#define PF_HELI_VEL_GOAL		11
#define PF_HELI_WIMP_FACTOR		12
#define PF_HELI_LAST_BDIST		13



void	get_bullet __STV_Prototype__(( int, STV_REGION,
			STV_PARTICLE_ID *, int * ));

void	get_bullet_pfield __STV_Prototype__(( STV_PARTICLE_ID,
			STV_VALUE, STV_VALUE * ));


void	make_bullet __STV_Prototype__(( int *, STV_PARTICLE_ID * ));

void	set_bullet_pfield __STV_Prototype__(( STV_PARTICLE_ID,
			STV_VALUE, STV_VALUE ));


#define PF_BULLET_BID			 0
#define PF_BULLET_AGE			 1
#define PF_BULLET_VEL_ANGLE		 2
#define PF_BULLET_VEL_MAGN		 3
#define PF_BULLET_PARENT		 4



void	get_explosion __STV_Prototype__(( int, STV_REGION,
			STV_PARTICLE_ID *, int * ));

void	get_explosion_pfield __STV_Prototype__(( STV_PARTICLE_ID,
			STV_VALUE, STV_VALUE * ));


void	make_explosion __STV_Prototype__(( int *, STV_PARTICLE_ID * ));

void	set_explosion_pfield __STV_Prototype__(( STV_PARTICLE_ID,
			STV_VALUE, STV_VALUE ));


#define PF_EXPLOSION_EID		 0
#define PF_EXPLOSION_STATE		 1
#define PF_EXPLOSION_ANGLE		 2
#define PF_EXPLOSION_STRENGTH	 3
#define PF_EXPLOSION_RADIUS		 4


REGION		create_region();

HELI		create_heli(), add_heli(), get_heli_by_hid();

BULLET		create_bullet(), add_bullet();

EXPLOSION	create_explosion(), add_explosion();

BLIP		create_blip(), add_blip();

double		HELI_HEAD_GOAL, HELI_VEL_GOAL, HELI_RADAR_RANGE;
int			HELI_SHIELD, HELI_FIRE, HELI_INDEX, HELI_RELEASE;

double		WIND_ANGLE, WIND_MAGN;

double		angle_of();
double		atof();
double		distance();
double		rand_angle();
double		rand_double();
double		sq();

void		add_local_blips();
void		add_rand_heli();
void		add_score();
void		bullet_handle();
void		check_exits();
void		cleanup_region();
void		do_stv_stuff();
void		eff_vel();
void		explosion_handle();
void		fire_bullet();
void		free_blip();
void		free_blip_list();
void		free_bullet();
void		free_explosion();
void		free_heli();
void		free_region();
void		free_region_list();
void		handle_steering();
void		heli_action();
void		heli_handle();
void		heli_init();
void		memcheck();
void		normalize_angle();
void		program_init();
void		pvm_init();
void		rand_point();
void		read_args();
void		region_init();
void		reset_steered();
void		send_transfers();
void		usage();
void		wind_effect();

int			NEXT_EXPLOSION_ID;

int			NEXT_BULLET_ID;

int			NEXT_HELI_ID;


/* Externals */

extern		int 	errno;


/* Global Variables */

REGION		GLOBREGION;

REGION		MYREGION;

HELI		HELIS;

BULLET		BULLETS;

EXPLOSION	EXPLOSIONS;

BLIP		BLIPS;

double		BULLET_VELOCITY;

double		HELI_VELOCITY;

double		VELOCITY_DELTA;
double		ANGLE_DELTA;
double		SPEED_DELTA;

double		WANDER_PROB;
double		FIRE_PROB;
double		HELI_PROB;

int			HELI_RADIUS;

char		**Argv;
int			Argc;

int			MYINST;
int			NTASKS;
int			MYTID;
int			SPMD;

int 		vflag;
int 		dflag;
int 		cflag;


/* Main */

main( argc, argv )
int argc;
char **argv;
{
	int cc;

	Argc = argc;
	Argv = argv;

	read_args();

	program_init();

	pvm_init();

	do_stv_stuff();

	heli_init();

	cc = stvTrue;

	while ( stvTrue )
	{
		if ( !cc )
		{
			cleanup_region();

			region_init();

			reset_steered();
		}

		cc = iterate();
	}
}


iterate()
{
	static int cnt = 0;

	int new_params;
	int cc;

	/* Dump Out Iteration Tick */

	if ( !(++cnt % 100) )
	{
		printf( "." );
		fflush( stdout );
	}

	/* Update Objects */

	explosion_handle();

	bullet_handle();

	heli_handle();

	/* Reset Radar Blips */

	free_blip_list( &BLIPS );

	add_local_blips();

	/* Communicate Transfers */

	send_transfers();

	check_exits();

	cc = recv_transfers();

	if ( !cc )
		return( stvFalse );

	/* Check for Actions */

	heli_action();

	/* Handle STV Requests */

	new_params = stv_sendToFE();

	if ( new_params )
		handle_steering();

	return( stvTrue );
}


void
heli_handle()
{
	EXPLOSION E;

	HELI Hlast;
	HELI Hnext;
	HELI H, H2;

	double factor1, factor2;
	double ra1, ra2;
	double a1, a2;
	double hdiff;
	double hmag;

	double hx, hy;

	int strength;
	int absorb;
	int dx, dy;
	int hit;

	/* Update Each Helicopter */

	H = HELIS;

	Hlast = (HELI) NULL;

	while ( H != NULL )
	{
		if ( H->killed )
		{
			/* Remove Helicopter */

			if ( vflag )
			{
				printf( "\n*** Killed Helicopter %d ***\n", H->hid );
				fflush( stdout );
			}

			add_explosion( H->loc_x, H->loc_y, H->vel_angle,
				HELI_EXPLOSION_STRENGTH );

			if ( Hlast == NULL )
				HELIS = H->next;

			else
				Hlast->next = H->next;

			Hnext = H->next;

			free_heli( &H );

			H = Hnext;
		}

		else
		{
			/* Check for Other Helicopter Collision */

			H2 = H->next;

			absorb = stvFalse;

			hit = stvFalse;

			while ( H2 != NULL && !hit )
			{
				if ( H->killed == stvFalse
					&& heli_hits_heli( H, H2, &hx, &hy ) )
				{
					if ( vflag )
					{
						printf( "\n*** Mid-Air Collision %d/%d ***\n",
							H->hid, H2->hid );
						fflush( stdout );
					}

					/* Determine Damage Strength */

					a1 = angle_of(
							(double) H->loc_x, (double) H->loc_y,
							hx, hy );

					a2 = angle_of(
							(double) H2->loc_x, (double) H2->loc_y,
							hx, hy );

					ra1 = a1 - H->vel_angle;
					ra2 = a2 - H2->vel_angle;

					factor1 = ( 1.0 + cos( ra1 ) ) / 4.0;
					factor2 = ( 1.0 + cos( ra2 ) ) / 4.0;

					strength = (int) ( ( H->vel_magn * factor1 )
						+ ( H2->vel_magn * factor2 ) ) / 2;

					/* Damage Helis */

					if ( !damage_heli( H, hx, hy, strength )
						&& !damage_heli( H2, hx, hy, strength ) )
					{
						/* Stick & Bounce Off */

						H->vel_angle = a1 - PI;
						H2->vel_angle = a2 - PI;

						if ( strength > 0 )
						{
							H->bounce = stvTrue;
							H2->bounce = stvTrue;
						}

						absorb = stvTrue;
					}

					hit = stvTrue;
				}

				else
					H2 = H2->next;
			}

			/* Check for Explosion Collision */

			E = EXPLOSIONS;

			while ( E != NULL && !hit )
			{
				if ( heli_hits_explosion( H, E, &hx, &hy ) )
				{
					if ( vflag )
					{
						printf(
							"\n*** Collision With Explosion %d ***\n",
							H->hid );
						fflush( stdout );
					}

					strength = E->radius / BULLET_EXPLOSION_STRENGTH;

					if ( !damage_heli( H, hx, hy, E->strength ) )
						absorb = stvTrue;

					hit = stvTrue;
				}

				else
					E = E->next;
			}

			if ( hit && !absorb )
			{
				/* Remove Helicopter */

				add_explosion( H->loc_x, H->loc_y, H->vel_angle,
					HELI_EXPLOSION_STRENGTH );

				if ( Hlast == NULL )
					HELIS = H->next;

				else
					Hlast->next = H->next;

				Hnext = H->next;

				free_heli( &H );

				H = Hnext;
			}

			else
			{
				/* Adjust Heading */

				if ( H->heading != H->head_goal )
				{
					hdiff = H->head_goal - H->heading;

					normalize_angle( &hdiff, &hmag );

					if ( hmag < SUPER_DELTA( ANGLE_DELTA, H ) )
						H->heading = H->head_goal;
			
					else if ( hdiff > 0 )
					{
						H->heading += SUPER_DELTA( ANGLE_DELTA, H );

						H->heading = H->heading > PI ?
							H->heading - ( 2.0 * PI ) : H->heading;
					}

					else
					{
						H->heading -= SUPER_DELTA( ANGLE_DELTA, H );

						H->heading = H->heading < -1.0 * PI ?
							H->heading + ( 2.0 * PI ) : H->heading;
					}
				}

				/* Adjust Velocity Heading */

				if ( H->vel_angle != H->heading )
				{
					hdiff = H->heading - H->vel_angle;

					normalize_angle( &hdiff, &hmag );

					if ( hmag < SUPER_DELTA( VELOCITY_DELTA, H ) )
						H->vel_angle = H->heading;
			
					else if ( hdiff > 0 )
					{
						H->vel_angle +=
							SUPER_DELTA( VELOCITY_DELTA, H );

						H->vel_angle = H->vel_angle > PI ?
							H->vel_angle - ( 2.0 * PI ) : H->vel_angle;
					}

					else
					{
						H->vel_angle -=
							SUPER_DELTA( VELOCITY_DELTA, H );

						H->vel_angle = H->vel_angle < -1.0 * PI ?
							H->vel_angle + ( 2.0 * PI ) : H->vel_angle;
					}
				}

				/* Adjust Velocity Magnitude */

				if ( H->vel_magn != SUPER_DELTA( H->vel_goal, H ) )
				{
					hdiff = SUPER_DELTA( H->vel_goal, H ) - H->vel_magn;

					hmag = fabs( hdiff );

					if ( hmag < SUPER_DELTA( SPEED_DELTA, H ) )
						H->vel_magn = SUPER_DELTA( H->vel_goal, H );
			
					else if ( hdiff > 0 )
						H->vel_magn += SUPER_DELTA( SPEED_DELTA, H );

					else
						H->vel_magn -= SUPER_DELTA( SPEED_DELTA, H );
				}

				/* Move Helicopter */

				if ( !(H->bounce) )
				{
					dx = (int) ( H->vel_magn
						* cos( H->vel_angle + (PI / 2.0) ) );

					dy = (int) ( H->vel_magn
						* sin( H->vel_angle + (PI / 2.0) ) );

					wind_effect( &dx, &dy );

					H->loc_x += dx;
					H->loc_y -= dy;
				}

				else
					H->bounce = stvFalse;

				/* Decrement Any Fire Delay */

				if ( H->fire_count > 0 )
					(H->fire_count)--;

				/* Next Helicopter */

				Hlast = H;

				H = H->next;
			}
		}
	}
}


void
bullet_handle()
{
	EXPLOSION E;

	BULLET Blast;
	BULLET Bnext;
	BULLET B, B2;

	HELI H;

	double hx, hy;

	int absorb;
	int dx, dy;
	int hit;

	/* Update Each Bullet */

	B = BULLETS;

	Blast = (BULLET) NULL;

	while ( B != NULL )
	{
		/* Check for Helicopter Collision */

		H = HELIS;

		absorb = stvFalse;

		hit = stvFalse;

		while ( H != NULL && !hit )
		{
			if ( B->parent != H->hid
				&& bullet_hits_heli( B, H, &hx, &hy ) )
			{
				if ( !damage_heli( H, hx, hy, 1 ) )
				{
					add_score( B->parent, BULLET_SCORE );

					absorb = stvTrue;
				}

				else
					add_score( B->parent, HELI_SCORE );

				hit = stvTrue;
			}

			else
				H = H->next;
		}

		/* Check for Other Bullet Collision */

		B2 = B->next;

		while ( B2 != NULL && !hit )
		{
			if ( B2->parent != B->parent
				&& bullet_hits_bullet( B, B2, &hx, &hy ) )
			{
				if ( B->age < B2->age )
					add_score( B->parent, BULLET_SCORE );
				
				else if ( B2->age < B->age )
					add_score( B2->parent, BULLET_SCORE );

				hit = stvTrue;
			}

			else
				B2 = B2->next;
		}

		/* Check for Explosion Collision */

		E = EXPLOSIONS;

		while ( E != NULL && !hit )
		{
			if ( bullet_hits_explosion( B, E, &hx, &hy ) )
				hit = stvTrue;

			else
				E = E->next;
		}

		/* Check for Death from Natural Causes :-) */

		(B->age)++;

		if ( !hit && B->age > BULLET_LIFETIME )
		{
			absorb = stvTrue;

			hit = stvTrue;
		}

		/* Move Bullet */

		if ( !hit )
		{
			dx = (int) ( B->vel_magn
				* cos( B->vel_angle + (PI / 2.0) ) );

			dy = (int) ( B->vel_magn
				* sin( B->vel_angle + (PI / 2.0) ) );

			wind_effect( &dx, &dy );

			B->loc_x += dx;
			B->loc_y -= dy;

			Blast = B;

			B = B->next;
		}

		/* Remove Bullet */

		else
		{
			if ( !absorb )
			{
				add_explosion( B->loc_x, B->loc_y, B->vel_angle,
					BULLET_EXPLOSION_STRENGTH );
			}

			if ( Blast == NULL )
				BULLETS = B->next;

			else
				Blast->next = B->next;

			Bnext = B->next;

			free_bullet( &B );

			B = Bnext;
		}
	}
}


void
wind_effect( dx, dy )
int *dx, *dy;
{
	int deltax, deltay;

	deltax = (int) ( WIND_MAGN * cos( WIND_ANGLE + (PI / 2.0) ) );

	deltay = (int) ( WIND_MAGN * sin( WIND_ANGLE + (PI / 2.0) ) );

	*dx += deltax;
	*dy += deltay;
}


bullet_hits_heli( B, H, gx, gy )
BULLET B;
HELI H;
double *gx, *gy;
{
	double eff_bvel_angle;
	double eff_hvel_angle;

	double eff_bvel_magn;
	double eff_hvel_magn;

	int hit;

	/* Determine Effective Velocity Angles & Magns */

	eff_vel( B->vel_magn, B->vel_angle,
		&eff_bvel_magn, &eff_bvel_angle );

	eff_vel( H->vel_magn, H->vel_angle,
		&eff_hvel_magn, &eff_hvel_angle );

	/* Check for Possibility of Collision */

	if ( not_even_close(
		B->loc_x, B->loc_y, eff_bvel_magn, BULLET_RADIUS,
		H->loc_x, H->loc_y, eff_hvel_magn, HELI_RADIUS ) )
	{
		*gx = *gy = -1.0;

		return( stvFalse );
	}

	hit = collision_detect(
		(double) B->loc_x, (double) B->loc_y,
			eff_bvel_angle, eff_bvel_magn, BULLET_RADIUS,
		(double) H->loc_x, (double) H->loc_y,
			eff_hvel_angle, eff_hvel_magn, HELI_RADIUS,
		gx, gy );
		
	return( hit );
}


bullet_hits_bullet( B, B2, gx, gy )
BULLET B, B2;
double *gx, *gy;
{
	double eff_bvel_angle1;
	double eff_bvel_angle2;

	double eff_bvel_magn1;
	double eff_bvel_magn2;

	int hit;

	/* Determine Effective Velocity Angles & Magns */

	eff_vel( B->vel_magn, B->vel_angle,
		&eff_bvel_magn1, &eff_bvel_angle1 );

	eff_vel( B2->vel_magn, B2->vel_angle,
		&eff_bvel_magn2, &eff_bvel_angle2 );

	/* Check for Possibility of Collision */

	if ( not_even_close(
		B->loc_x, B->loc_y, eff_bvel_magn1, BULLET_RADIUS,
		B2->loc_x, B2->loc_y, eff_bvel_magn2, BULLET_RADIUS ) )
	{
		*gx = *gy = -1.0;

		return( stvFalse );
	}

	hit = collision_detect(
		(double) B->loc_x, (double) B->loc_y,
			eff_bvel_angle1, eff_bvel_magn1, BULLET_RADIUS,
		(double) B2->loc_x, (double) B2->loc_y,
			eff_bvel_angle2, eff_bvel_magn2, BULLET_RADIUS,
		gx, gy );
		
	return( hit );
}


heli_hits_heli( H, H2, gx, gy )
HELI H, H2;
double *gx, *gy;
{
	double eff_hvel_angle1;
	double eff_hvel_angle2;

	double eff_hvel_magn1;
	double eff_hvel_magn2;

	int hit;

	/* Determine Effective Velocity Angles & Magns */

	eff_vel( H->vel_magn, H->vel_angle,
		&eff_hvel_magn1, &eff_hvel_angle1 );

	eff_vel( H2->vel_magn, H2->vel_angle,
		&eff_hvel_magn2, &eff_hvel_angle2 );

	/* Check for Possibility of Collision */

	if ( not_even_close(
		H->loc_x, H->loc_y, eff_hvel_magn1, HELI_RADIUS,
		H2->loc_x, H2->loc_y, eff_hvel_magn2, HELI_RADIUS ) )
	{
		*gx = *gy = -1.0;

		return( stvFalse );
	}

	hit = collision_detect(
		(double) H->loc_x, (double) H->loc_y,
			eff_hvel_angle1, eff_hvel_magn1, HELI_RADIUS,
		(double) H2->loc_x, (double) H2->loc_y,
			eff_hvel_angle2, eff_hvel_magn2, HELI_RADIUS,
		gx, gy );
		
	return( hit );
}


bullet_hits_explosion( B, E, gx, gy )
BULLET B;
EXPLOSION E;
double *gx, *gy;
{
	double eff_bvel_angle;
	double eff_evel_angle;

	double eff_bvel_magn;
	double eff_evel_magn;

	int hit;

	/* Determine Effective Velocity Angles & Magns */

	eff_vel( B->vel_magn, B->vel_angle,
		&eff_bvel_magn, &eff_bvel_angle );

	eff_vel( (double) 0.0, (double) 0.0,
		&eff_evel_magn, &eff_evel_angle );

	/* Check for Possibility of Collision */

	if ( not_even_close(
		B->loc_x, B->loc_y, eff_bvel_magn, BULLET_RADIUS,
		E->loc_x, E->loc_y, eff_evel_magn, E->radius ) )
	{
		*gx = *gy = -1.0;

		return( stvFalse );
	}

	if ( inside_explosion( B->loc_x, B->loc_y, E ) )
	{
		*gx = B->loc_x;
		*gy = B->loc_y;

		return( stvTrue );
	}

	hit = collision_detect(
		(double) B->loc_x, (double) B->loc_y,
			eff_bvel_angle, eff_bvel_magn, BULLET_RADIUS,
		(double) E->loc_x, (double) E->loc_y,
			eff_evel_angle, eff_evel_magn, E->radius,
		gx, gy );
		
	return( hit );
}


heli_hits_explosion( H, E, gx, gy )
HELI H;
EXPLOSION E;
double *gx, *gy;
{
	double eff_hvel_angle;
	double eff_evel_angle;

	double eff_hvel_magn;
	double eff_evel_magn;

	int hit;

	/* Determine Effective Velocity Angles & Magns */

	eff_vel( H->vel_magn, H->vel_angle,
		&eff_hvel_magn, &eff_hvel_angle );

	eff_vel( (double) 0.0, (double) 0.0,
		&eff_evel_magn, &eff_evel_angle );

	/* Check for Possibility of Collision */

	if ( not_even_close(
		H->loc_x, H->loc_y, eff_hvel_magn, HELI_RADIUS,
		E->loc_x, E->loc_y, eff_evel_magn, E->radius ) )
	{
		*gx = *gy = -1.0;

		return( stvFalse );
	}

	if ( inside_explosion( H->loc_x, H->loc_y, E ) )
	{
		*gx = H->loc_x;
		*gy = H->loc_y;

		return( stvTrue );
	}

	hit = collision_detect(
		(double) H->loc_x, (double) H->loc_y,
			eff_hvel_angle, eff_hvel_magn, HELI_RADIUS,
		(double) E->loc_x, (double) E->loc_y,
			eff_evel_angle, eff_evel_magn, E->radius,
		gx, gy );
		
	return( hit );
}


inside_explosion( x, y, E )
int x, y;
EXPLOSION E;
{
	double dist;

	dist = distance( (double) x, (double) y,
		(double) E->loc_x, (double) E->loc_y );

	if ( dist <= (double) E->radius )
		return( stvTrue );
	
	else
		return( stvFalse );
}


void
eff_vel( magn, angle, eff_magn, eff_angle )
double magn, angle;
double *eff_magn, *eff_angle;
{
	double x, y;

	x = magn * cos( angle + (PI / 2.0) );
	y = magn * sin( angle + (PI / 2.0) );

	x += WIND_MAGN * cos( WIND_ANGLE + (PI / 2.0) );
	y += WIND_MAGN * sin( WIND_ANGLE + (PI / 2.0) );

	*eff_magn = sqrt( ( x * x ) + ( y * y ) );

	*eff_angle = angle_of( (double) 0.0, (double) 0.0, x, -1.0 * y );
}


void
explosion_handle()
{
	EXPLOSION Elast;
	EXPLOSION Enext;
	EXPLOSION E;

	int dx, dy;
	int delta;

	/* Update Each Explosion */

	E = EXPLOSIONS;

	Elast = (EXPLOSION) NULL;

	while ( E != NULL )
	{
		dx = dy = 0;

		wind_effect( &dx, &dy );

		E->loc_x += dx;
		E->loc_y -= dy;

		if ( E->state == EXPLOSION_GROWING )
		{
			delta = ( E->strength - E->radius ) / 2;

			delta = delta == 0 ? 1 : delta;

			E->radius += delta;

			if ( E->radius == E->strength )
				E->state = EXPLOSION_SHRINKING;

			Elast = E;

			E = E->next;
		}

		else if ( E->state == EXPLOSION_SHRINKING )
		{
			(E->radius)--;

			if ( !(E->radius) )
			{
				DUMP_EXPLOSION( "removing", E->eid, E->loc_x, E->loc_y,
					E->angle, E->strength );

				if ( Elast == NULL )
					EXPLOSIONS = E->next;

				else
					Elast->next = E->next;

				Enext = E->next;

				free_explosion( &E );

				E = Enext;
			}

			else
			{
				Elast = E;

				E = E->next;
			}
		}
	}
}


not_even_close( x1, y1, vel1, rad1, x2, y2, vel2, rad2 )
int x1, y1;
double vel1;
int rad1;
int x2, y2;
double vel2;
int rad2;
{
	double dist;

	dist = distance( (double) x1, (double) y1,
		(double) x2, (double) y2 );

	if ( dist > vel1 + (double) rad1 + vel2 + (double) rad2 + 1.0 )
		return( stvTrue );
	
	else
		return( stvFalse );
}


collision_detect(
	x10, y10, angle1, magn1, rad1,
	x20, y20, angle2, magn2, rad2,
	xh, yh )
double x10, y10;
double angle1;
double magn1;
int rad1;
double x20, y20;
double angle2;
double magn2;
int rad2;
double *xh, *yh;
{
	double B1, B2, C1, G, H1, H2, I1, J1;
	double A, B, C;
	double DET;
	double R;

	double theta1, theta2;
	double a1, b1;
	double a2, b2;
	double r1, r2;

	double a1sq;
	double a2sq;

	double pdiff, mdiff;
	double x1p, x1m;
	double x2p, x2m;
	double y1p, y1m;
	double x1, y1;
	double x2, y2;

	double dck1, dck2;

	y10 = ( (double) FIELD_HEIGHT ) - y10;
	y20 = ( (double) FIELD_HEIGHT ) - y20;

	theta1 = angle1 + ( PI / 2.0 );
	theta2 = angle2 + ( PI / 2.0 );

	a1 = magn1 * cos( theta1 );
	b1 = magn1 * sin( theta1 );

	a2 = magn2 * cos( theta2 );
	b2 = magn2 * sin( theta2 );

	r1 = (double) rad1;
	r2 = (double) rad2;

	if ( cflag )
	{
		printf( "\n[(%lf,%lf):%lf@%lf->%lf,%lf,%lf,%lf]\n",
			x10, y10, magn1, angle1, theta1, a1, b1, r1 );

		printf( "[(%lf,%lf):%lf@%lf->%lf,%lf,%lf,%lf]\n",
			x20, y20, magn2, angle2, theta2, a2, b2, r2 );
	}

	if ( EQUIV( a1, 0.0 ) && EQUIV( a2, 0.0 ) )
	{
		I1 = y20 - ( ( b2 * y10 ) / b1 );

		J1 = ( b2 / b1 ) - 1;

		R = r1 + r2;

		A = sq( J1 );

		B = 2.0 * I1 * J1;

		C = sq( I1 ) + sq( x10 - x20 ) - sq( R );

		DET = sq( B ) - ( 4.0 * A * C );

		/* Check For No Intersection - MISS */

		if ( DET < 0 )
		{
			if ( cflag )
				{ printf( "\tDETA\n" ); fflush( stdout ); }

			*xh = *yh = -1.0;

			return( stvFalse );
		}

		/* Determine Correct Solution Point */

		y1p = ( ( -1.0 * B ) + sqrt( DET ) ) / ( 2.0 * A );
		y1m = ( ( -1.0 * B ) - sqrt( DET ) ) / ( 2.0 * A );

		pdiff = fabs( y1p - y10 );

		mdiff = fabs( y1m - y10 );

		if ( pdiff < mdiff )
			y1 = y1p;

		else
			y1 = y1m;

		/* Get Rest of Object Coords */

		x1 = x10;

		x2 = x20;

		y2 = ( ( b2 * y1 ) / b1 ) + I1;
	}

	else if ( EQUIV( a1, 0.0 ) && !EQUIV( a2, 0.0 ) )
	{
		a2sq = sq( a2 );

		B2 = y20 - ( ( b2 * x20 ) / a2 );

		H2 = y10 - ( ( b1 * x20 ) / a2 );

		R = r1 + r2;

		A = ( sq( a2 ) + sq( b2 - b1 ) ) / a2sq;

		B = ( 2.0
			* ( ( ( b2 - b1 ) * ( B2 - H2 ) ) - ( a2 * x10 ) ) )
			/ a2;

		C = sq( B2 - H2 ) + sq( x10 ) - sq( R );

		DET = sq( B ) - ( 4.0 * A * C );

		/* Check For No Intersection - MISS */

		if ( DET < 0 )
		{
			if ( cflag )
				{ printf( "\tDETB\n" ); fflush( stdout ); }

			*xh = *yh = -1.0;

			return( stvFalse );
		}

		/* Determine Correct Solution Point */

		x2p = ( ( -1.0 * B ) + sqrt( DET ) ) / ( 2.0 * A );
		x2m = ( ( -1.0 * B ) - sqrt( DET ) ) / ( 2.0 * A );

		pdiff = fabs( x2p - x20 );

		mdiff = fabs( x2m - x20 );

		if ( pdiff < mdiff )
			x2 = x2p;

		else
			x2 = x2m;

		/* Get Rest of Object Coords */

		x1 = x10;

		y1 = ( ( b1 * x2 ) / a2 ) + H2;

		y2 = ( ( b2 * x2 ) / a2 ) + B2;
	}

	else if ( !EQUIV( a1, 0.0 ) && EQUIV( a2, 0.0 ) )
	{
		a1sq = sq( a1 );

		B1 = y10 - ( ( b1 * x10 ) / a1 );

		H1 = y20 - ( ( b2 * x10 ) / a1 );

		R = r1 + r2;

		A = ( sq( a1 ) + sq( b2 - b1 ) ) / a1sq;

		B = ( 2.0
			* ( ( ( b1 - b2 ) * ( B1 - H1 ) ) - ( a1 * x20 ) ) )
			/ a1;

		C = sq( B1 - H1 ) + sq( x20 ) - sq( R );

		DET = sq( B ) - ( 4.0 * A * C );

		/* Check For No Intersection - MISS */

		if ( DET < 0 )
		{
			if ( cflag )
				{ printf( "\tDETC\n" ); fflush( stdout ); }

			*xh = *yh = -1.0;

			return( stvFalse );
		}

		/* Determine Correct Solution Point */

		x1p = ( ( -1.0 * B ) + sqrt( DET ) ) / ( 2.0 * A );
		x1m = ( ( -1.0 * B ) - sqrt( DET ) ) / ( 2.0 * A );

		pdiff = fabs( x1p - x10 );

		mdiff = fabs( x1m - x10 );

		if ( pdiff < mdiff )
			x1 = x1p;

		else
			x1 = x1m;

		/* Get Rest of Object Coords */

		x2 = x20;

		y2 = ( ( b2 * x1 ) / a1 ) + H1;

		y1 = ( ( b1 * x1 ) / a1 ) + B1;
	}

	else
	{
		a1sq = sq( a1 );

		B1 = y10 - ( ( b1 * x10 ) / a1 );
		B2 = y20 - ( ( b2 * x20 ) / a2 );

		C1 = x20 - ( ( a2 * x10 ) / a1 );

		G = ( ( b2 * C1 ) / a2 ) + B2;

		R = r1 + r2;

		A = ( sq( a2 - a1 ) + sq( b2 - b1 ) ) / a1sq;

		B = ( 2.0
			* ( ( C1 * ( a2 - a1 ) ) + ( ( G - B1 ) * ( b2 - b1 ) ) ) )
			/ a1;

		C = sq( B1 - G ) + sq( C1 ) - sq( R );

		DET = sq( B ) - ( 4.0 * A * C );

		/* Check For No Intersection - MISS */

		if ( DET < 0 )
		{
			if ( cflag )
				{ printf( "\tDETD\n" ); fflush( stdout ); }

			*xh = *yh = -1.0;

			return( stvFalse );
		}

		/* Determine Correct Solution Point */

		x1p = ( ( -1.0 * B ) + sqrt( DET ) ) / ( 2.0 * A );
		x1m = ( ( -1.0 * B ) - sqrt( DET ) ) / ( 2.0 * A );

		pdiff = fabs( x1p - x10 );

		mdiff = fabs( x1m - x10 );

		if ( pdiff < mdiff )
			x1 = x1p;

		else
			x1 = x1m;

		/* Get Rest of Object Coords */

		y1 =  (( b1 * x1 ) / a1 ) + B1;

		x2 = ( ( a2 * x1 ) / a1 ) + C1;

		y2 = ( ( b2 * x2 ) / a2 ) + B2;
	}

	/* Verify Coords Reached in Time */

	if ( !IN_RANGE( x1, x10, a1 )
		|| !IN_RANGE( y1, y10, b1 )
		|| !IN_RANGE( x2, x20, a2 )
		|| !IN_RANGE( y2, y20, b2 ) )
	{
		if ( cflag )
		{
			printf( "\tNIR\n" );

			printf(
	"\t\t%lf,%lf,%lf\n\t%lf,%lf,%lf\n\t%lf,%lf,%lf\n\t%lf,%lf,%lf\n",
				x1, x10, a1, y1, y10, b1, x2, x20, a2, y2, y20, b2 );
		}

		*xh = *yh = -1.0;

		return( stvFalse );
	}

	/* Calculate Glance Point */

	if ( cflag )
	{
		printf( "\t***\n" );

		printf(
		"\t\t(%lf,%lf,%lf/%lf,%lf,%lf/%lf,%lf,%lf/%lf,%lf,%lf)\n",
			x1, x10, a1, y1, y10, b1, x2, x20, a2, y2, y20, b2 );
	}

	y1 = ( (double) FIELD_HEIGHT ) - y1;
	y2 = ( (double) FIELD_HEIGHT ) - y2;

	if ( !EQUIV( R, 0.0 ) )
	{
		*xh = ( ( r1 * x2 ) + ( r2 * x1 ) ) / R;

		*yh = ( ( r1 * y2 ) + ( r2 * y1 ) ) / R;
	}

	else
	{
		*xh = x1;
		*yh = y1;
	}

	if ( cflag )
		printf( "\t\tH( %lf, %lf )\n", *xh, *yh );

	return( stvTrue );
}


void
add_local_blips()
{
	BULLET B;

	HELI H;

	int x, y;

	/* Add Heli Blips */

	H = HELIS;

	while ( H != NULL )
	{
		add_blip( H->loc_x, H->loc_y, (double) HELI_RADIUS );

		H = H->next;
	}

	/* Add Bullet Blips */

	B = BULLETS;

	while ( B != NULL )
	{
		add_blip( B->loc_x, B->loc_y, (double) BULLET_RADIUS );

		B = B->next;
	}

	/* Add Some Random Blip Noise */

	while ( rand_double() < NOISE_LEVEL )
	{
		rand_point( MYREGION, &x, &y );

		add_blip( x, y, (double) BULLET_RADIUS );
	}
}


void
add_score( hid, score )
int hid;
int score;
{
	HELI H;

	int tmp;

	H = get_heli_by_hid( hid );

	if ( H != NULL )
		H->score += score;

	else
	{
		pvm_initsend(PvmDataDefault);

		tmp = SCORE_TRANSFER;
		pvm_pkint( &tmp, 1, 1 );

		pvm_pkint( &hid, 1, 1 );

		pvm_pkint( &score, 1, 1 );

		tmp = END_TRANSFER;
		pvm_pkint( &tmp, 1, 1 );

		pvm_bcast( GROUPNAME, TRANSFER_MSG );
	}
}


void
send_transfers()
{
	REGION R;

	BULLET Blast;
	BULLET Bnext;
	BULLET B;

	HELI Hlast;
	HELI Hnext;
	HELI H;

	double radius;

	int tmp;

	/* Check for Transfers to Each Neighbor */

	R = MYREGION->neigh;

	while ( R != NULL )
	{
		/* Initialize Message */

		pvm_initsend( PvmDataDefault );

		/* Check for Bullets */

		B = BULLETS;

		Blast = (BULLET) NULL;

		while ( B != NULL )
		{
			if ( B->loc_x >= R->xmin && B->loc_x <= R->xmax
				&& B->loc_y >= R->ymin && B->loc_y <= R->ymax )
			{
				/* Transfer Bullet */

				DUMP_BULLET( "\nsending", B->bid, B->loc_x, B->loc_y,
					B->vel_angle, B->vel_magn );

				tmp = BULLET_TRANSFER;
				pvm_pkint( &tmp, 1, 1 );

				pvm_pkint( &(B->bid), 1, 1 );
				pvm_pkint( &(B->age), 1, 1 );
				pvm_pkint( &(B->loc_x), 1, 1 );
				pvm_pkint( &(B->loc_y), 1, 1 );
				pvm_pkdouble( &(B->vel_angle), 1, 1 );
				pvm_pkdouble( &(B->vel_magn), 1, 1 );
				pvm_pkint( &(B->parent), 1, 1 );

				/* Remove From List */

				if ( Blast == NULL )
					BULLETS = B->next;

				else
					Blast->next = B->next;

				Bnext = B->next;

				free_bullet( &B );

				B = Bnext;
			}

			else
			{
				/* Send Radar Blip */

				tmp = BLIP_TRANSFER;
				pvm_pkint( &tmp, 1, 1 );

				pvm_pkint( &(B->loc_x), 1, 1 );
				pvm_pkint( &(B->loc_y), 1, 1 );

				radius = (double) BULLET_RADIUS;
				pvm_pkdouble( &(radius), 1, 1 );

				Blast = B;

				B = B->next;
			}
		}

		/* Check for Helicopters */

		H = HELIS;

		Hlast = (HELI) NULL;

		while ( H != NULL )
		{
			if ( H->loc_x >= R->xmin && H->loc_x <= R->xmax
				&& H->loc_y >= R->ymin && H->loc_y <= R->ymax )
			{
				/* Transfer Helicopter */

				DUMP_HELI( "\nsending", H->hid, H->loc_x, H->loc_y,
					H->radar_range, H->heading, H->head_goal,
					H->vel_angle, H->vel_magn, H->vel_goal );

				tmp = HELI_TRANSFER;
				pvm_pkint( &tmp, 1, 1 );

				pvm_pkint( &(H->hid), 1, 1 );
				pvm_pkint( &(H->loc_x), 1, 1 );
				pvm_pkint( &(H->loc_y), 1, 1 );
				pvm_pkint( &(H->fire_count), 1, 1 );
				pvm_pkint( &(H->steered), 1, 1 );
				pvm_pkint( &(H->shield_str), 1, 1 );
				pvm_pkint( &(H->shielded), 1, 1 );
				pvm_pkint( &(H->score), 1, 1 );
				pvm_pkdouble( &(H->radar_range), 1, 1 );
				pvm_pkdouble( &(H->head_goal), 1, 1 );
				pvm_pkdouble( &(H->heading), 1, 1 );
				pvm_pkdouble( &(H->vel_angle), 1, 1 );
				pvm_pkdouble( &(H->vel_magn), 1, 1 );
				pvm_pkdouble( &(H->vel_goal), 1, 1 );
				pvm_pkdouble( &(H->wimp_factor), 1, 1 );
				pvm_pkdouble( &(H->last_bdist), 1, 1 );

				/* Remove From List */

				if ( Hlast == NULL )
					HELIS = H->next;

				else
					Hlast->next = H->next;

				Hnext = H->next;

				free_heli( &H );

				H = Hnext;
			}

			else
			{
				/* Send Radar Blip */

				tmp = BLIP_TRANSFER;
				pvm_pkint( &tmp, 1, 1 );

				pvm_pkint( &(H->loc_x), 1, 1 );
				pvm_pkint( &(H->loc_y), 1, 1 );

				radius = (double) HELI_RADIUS;
				pvm_pkdouble( &(radius), 1, 1 );

				Hlast = H;

				H = H->next;
			}
		}

		tmp = END_TRANSFER;
		pvm_pkint( &tmp, 1, 1 );

		pvm_send( R->tid, TRANSFER_MSG );

		R = R->neigh;
	}
}


void
check_exits()
{
	BULLET Blast;
	BULLET Bnext;
	BULLET B;

	HELI Hlast;
	HELI Hnext;
	HELI H;

	/* Check for Bullet Exits */

	B = BULLETS;

	Blast = (BULLET) NULL;

	while ( B != NULL )
	{
		if ( B->loc_x < MYREGION->xmin || B->loc_x > MYREGION->xmax
			|| B->loc_y < MYREGION->ymin || B->loc_y > MYREGION->ymax )
		{
			/* Remove From List */

			DUMP_BULLET( "\nexit", B->bid, B->loc_x, B->loc_y,
				B->vel_angle, B->vel_magn );

			if ( Blast == NULL )
				BULLETS = B->next;

			else
				Blast->next = B->next;

			Bnext = B->next;

			free_bullet( &B );

			B = Bnext;
		}

		else
		{
			Blast = B;

			B = B->next;
		}
	}

	/* Check for Helicopter Exits */

	H = HELIS;

	Hlast = (HELI) NULL;

	while ( H != NULL )
	{
		if ( H->loc_x < MYREGION->xmin || H->loc_x > MYREGION->xmax
			|| H->loc_y < MYREGION->ymin || H->loc_y > MYREGION->ymax )
		{
			/* Remove From List */

			DUMP_HELI( "\nexit", H->hid, H->loc_x, H->loc_y,
				H->radar_range, H->heading, H->head_goal,
				H->vel_angle, H->vel_magn, H->vel_goal );

			if ( Hlast == NULL )
				HELIS = H->next;

			else
				Hlast->next = H->next;

			Hnext = H->next;

			free_heli( &H );

			H = Hnext;
		}

		else
		{
			Hlast = H;

			H = H->next;
		}
	}
}


int
recv_transfers()
{
	REGION R;

	HELI H;

	double radar_range;

	double head_goal;
	double heading;

	double vel_angle;
	double vel_magn;
	double vel_goal;

	double wimp_factor;
	double last_bdist;

	double radius;

	int shield_str;
	int fire_count;
	int shielded;
	int steered;
	int msgtag;
	int parent;
	int score;
	int x, y;
	int buf;
	int bid;
	int age;
	int cnt;
	int hid;
	int tmp;

	/* Receive Transfers Back from Each Neighbor */

	if ( MYREGION->nneigh == 0 )
		return( stvTrue );

	/* Receive Transfers */

	cnt = 0;

	do
	{
		buf = pvm_recv( -1, TRANSFER_MSG );

		if ( buf < 0 )
		{
			pvm_perror( "Receiving Transfer Message" );

			return( stvFalse );
		}

		/* Transfer Message Received */

		if ( buf > 0 )
		{
			cnt++;

			/* Get Transfer Op */

			pvm_upkint( &tmp, 1, 1 );

			while ( tmp != END_TRANSFER )
			{
				switch ( tmp )
				{
					case BLIP_TRANSFER:
					{
						pvm_upkint( &x, 1, 1 );
						pvm_upkint( &y, 1, 1 );
						pvm_upkdouble( &radius, 1, 1 );

						add_blip( x, y, radius );

						break;
					}

					case BULLET_TRANSFER:
					{
						pvm_upkint( &bid, 1, 1 );
						pvm_upkint( &age, 1, 1 );
						pvm_upkint( &x, 1, 1 );
						pvm_upkint( &y, 1, 1 );
						pvm_upkdouble( &vel_angle, 1, 1 );
						pvm_upkdouble( &vel_magn, 1, 1 );
						pvm_upkint( &parent, 1, 1 );

						if ( vflag )
							printf( "\nbullet recvd:\n" );

						add_bullet( bid, age, x, y,
							vel_angle, vel_magn, parent );

						break;
					}

					case HELI_TRANSFER:
					{
						pvm_upkint( &hid, 1, 1 );
						pvm_upkint( &x, 1, 1 );
						pvm_upkint( &y, 1, 1 );
						pvm_upkint( &fire_count, 1, 1 );
						pvm_upkint( &steered, 1, 1 );
						pvm_upkint( &shield_str, 1, 1 );
						pvm_upkint( &shielded, 1, 1 );
						pvm_upkint( &score, 1, 1 );
						pvm_upkdouble( &radar_range, 1, 1 );
						pvm_upkdouble( &head_goal, 1, 1 );
						pvm_upkdouble( &heading, 1, 1 );
						pvm_upkdouble( &vel_angle, 1, 1 );
						pvm_upkdouble( &vel_magn, 1, 1 );
						pvm_upkdouble( &vel_goal, 1, 1 );
						pvm_upkdouble( &wimp_factor, 1, 1 );
						pvm_upkdouble( &last_bdist, 1, 1 );

						if ( vflag )
							printf( "\nheli recvd:\n" );

						add_heli( hid, x, y, fire_count, steered,
							shield_str, shielded,
							score, radar_range, head_goal, heading,
							vel_angle, vel_magn, vel_goal,
							wimp_factor, last_bdist );

						break;
					}

					case SCORE_TRANSFER:
					{
						pvm_upkint( &hid, 1, 1 );

						pvm_upkint( &score, 1, 1 );

						H = get_heli_by_hid( hid );

						if ( H != NULL )
							H->score += score;

						cnt--;

						break;
					}

					default:
					{
						if ( tmp > 0 )
						{
							printf( "\nTask %d/0x%x Died.  ",
								tmp, tmp );
							printf( "I'm outta here...\n\n" );

							exit( 0 );
						}

						else
						{
							printf( "Unknown Message Op = %d\n",
								tmp );
						}
					}
				}

				/* Get Next Transfer Op */

				pvm_upkint( &tmp, 1, 1 );
			}
		}
	}
	while ( cnt < MYREGION->nneigh );

	return( stvTrue );
}


void
heli_action()
{
	BLIP BH, BB, BE;
	BLIP BP;

	HELI H;

	double bullet_close;
	double heli_close;
	double exp_close;
	double dist;
	double r;

	double langle, rangle;
	double ldiff, rdiff;
	double lmag, rmag;
	double wimpout;
	double angle;
	double delta;

	double hdiff;
	double hmag;

	int flag;

	/* Check Each Helicopter */

	H = HELIS;

	while ( H != NULL )
	{
		if ( !(H->steered) )
		{
			/* Check Radar for Other Targets */

			BH = (BLIP) NULL;
			BB = (BLIP) NULL;
			BE = (BLIP) NULL;

			bullet_close = -1.0;
			heli_close = -1.0;
			exp_close = -1.0;

			BP = BLIPS;

			while ( BP != NULL )
			{
				dist = distance( (double) H->loc_x, (double) H->loc_y,
					(double) BP->loc_x, (double) BP->loc_y );

				/* Not Me, But Visible */

				if ( dist > 0.0 && dist < H->radar_range )
				{
					if ( EQUIV( BP->radius, (double) HELI_RADIUS ) )
					{
						if ( dist < heli_close || heli_close == -1.0 )
						{
							BH = BP;

							heli_close = dist;
						}
					}

					else if ( EQUIV( BP->radius,
						(double) BULLET_RADIUS ) )
					{
						if ( dist < bullet_close
							|| bullet_close == -1.0 )
						{
							BB = BP;

							bullet_close = dist;
						}
					}

					else
					{
						if ( dist < exp_close
							|| exp_close == -1.0 )
						{
							BE = BP;

							exp_close = dist;
						}
					}
				}

				BP = BP->next;
			}

			/* Reset Shield & Activity Flag */

			H->shielded = stvFalse;

			flag = 0;

			/* Set Wimp-Out Factor */

			wimpout = 40.0 * HELI_VELOCITY * H->wimp_factor;

			wimpout = wimpout < 10.0 * HELI_VELOCITY ?
				10.0 * HELI_VELOCITY : wimpout;

			/* Move Towards Closest Heli Blip */

			if ( BH != NULL )
			{
				/* Play Chicken */

				if ( heli_close < wimpout )
				{
					H->head_goal = angle_of(
						(double) BH->loc_x, (double) BH->loc_y,
						(double) H->loc_x, (double) H->loc_y );

					H->shielded = stvTrue;
				}

				else
				{
					H->head_goal = angle_of(
						(double) H->loc_x, (double) H->loc_y,
						(double) BH->loc_x, (double) BH->loc_y );

					/* Fire If Remotely On Target */

					hdiff = H->head_goal - H->heading;

					normalize_angle( &hdiff, &hmag );

					if ( hmag < PI / 6.0 )
						fire_bullet( H, "Blip" );
				}

				flag++;
			}

			/* Possible Evasive Maneuvers */

			if ( BB != NULL )
			{
				/* Possible Approaching Bullet */

				if ( H->last_bdist >= 0.0
					&& bullet_close < H->last_bdist
					&& H->last_bdist - bullet_close
						< 2.0 * BULLET_VELOCITY )
				{
					if ( bullet_close
						< 50.0 * BULLET_VELOCITY * H->wimp_factor )
					{
						angle = angle_of(
							(double) H->loc_x, (double) H->loc_y,
							(double) BB->loc_x, (double) BB->loc_y );

						langle = angle + ( PI / 2.0 )
							+ ( H->wimp_factor * PI / 2.0 )
							+ ( rand_angle() / 10.0 );
						ldiff = langle - H->heading;
						normalize_angle( &ldiff, &lmag );

						rangle = angle - ( PI / 2.0 )
							- ( H->wimp_factor * PI / 2.0 )
							+ ( rand_angle() / 10.0 );
						rdiff = rangle - H->heading;
						normalize_angle( &rdiff, &rmag );

						if ( lmag < rmag )
							H->head_goal = langle;
			
						else
							H->head_goal = rangle;

						flag++;
					}

					if ( bullet_close
						< 6.0 * BULLET_VELOCITY * H->wimp_factor )
					{
						H->shielded = stvTrue;

						flag++;
					}
				}
			}

			H->last_bdist = bullet_close;

			/* Avoid Possible Explosions */

			if ( BE != NULL )
			{
				/* Run Away! */

				if ( exp_close < wimpout )
				{
					H->head_goal = angle_of(
						(double) BE->loc_x, (double) BE->loc_y,
						(double) H->loc_x, (double) H->loc_y );

					H->shielded = stvTrue;

					flag++;
				}
			}

			/* Otherwise, Wander Aimlessly */

			if ( !flag )
			{
				if ( rand_double() < WANDER_PROB )
				{
					H->head_goal += ( -1.0 * PI / 4.0 )
						+ ( ( PI / 2.0 ) * rand_double() );
				}

				/* Check for Random Shot-in-the-Dark Firing */

				if ( rand_double() < FIRE_PROB )
					fire_bullet( H, "Random" );
			}
		}

		H = H->next;
	}

	/* Check for Helicopter Additions */

	if ( HELIS == NULL )
	{
		if ( rand_double() < 2.0 * HELI_PROB )
			add_rand_heli();
	}

	else
	{
		if ( rand_double() < HELI_PROB )
			add_rand_heli();
	}
}


void
handle_steering()
{
	HELI H;

	do
	{
		/* Handle Helicopter Steering */

		H = get_heli_by_hid( HELI_INDEX );

		if ( H != NULL )
		{
			if ( HELI_RELEASE == stvTrue )
				H->steered = stvFalse;
			
			else
			{
				H->steered = stvTrue;

				H->head_goal = HELI_HEAD_GOAL;

				H->vel_goal = HELI_VEL_GOAL;

				H->radar_range = HELI_RADAR_RANGE;

				H->shielded = HELI_SHIELD;

				if ( HELI_FIRE == stvTrue )
					fire_bullet( H, "steered" );
			}
		}

		/* Reset Parameters */

		HELI_HEAD_GOAL = HELI_VEL_GOAL = HELI_RADAR_RANGE = -1.0;

		HELI_SHIELD = HELI_FIRE = HELI_INDEX = HELI_RELEASE = -1;
	}
	while ( stv_get_remaining_params() > 0 );
}


void
read_args()
{
	int i, j, k;
	int do_usage;
	int len;

	HELI_VELOCITY = 30.0;

	BULLET_VELOCITY = HELI_VELOCITY * (double) SPEED_FACTOR;

	VELOCITY_DELTA = 0.060;

	ANGLE_DELTA = 0.100;

	SPEED_DELTA = 0.50;

	WANDER_PROB = 0.01;

	FIRE_PROB = 0.00005;

	HELI_PROB = 0.00001;

	HELI_RADIUS = 32;

	WIND_ANGLE = 0.0;

	WIND_MAGN = 0.0;

	SPMD = stvFalse;

	NTASKS = 1;

	vflag = 0;
	dflag = 0;
	cflag = 0;

	stv_setopt( stvOptNull, stvOptVerbose, vflag ? STV_DEBUG_ALL : 0 );

	do_usage = 0;

	for ( i=0 ; i < Argc ; i++ )
	{
		k = i + 1;

		if ( !strcmp( Argv[i], "-SPMD" ) )
			SPMD = stvTrue;

		else if ( !strcmp( Argv[i], "-WA" ) )
			WIND_ANGLE = atof( Argv[k++] );

		else if ( !strcmp( Argv[i], "-WM" ) )
			WIND_MAGN = atof( Argv[k++] );

		else if ( Argv[i][0] == '-' )
		{
			len = strlen( Argv[i] );

			for ( j=1 ; j < len ; j++ )
			{
				switch ( Argv[i][j] )
				{
					case 'N':
					{
						NTASKS = atoi( Argv[k++] );

						break;
					}

					case 'H':
					{
						HELI_VELOCITY = atof( Argv[k++] );

						BULLET_VELOCITY = HELI_VELOCITY
							* (double) SPEED_FACTOR;

						break;
					}

					case 'B':
					{
						BULLET_VELOCITY = atof( Argv[k++] );

						break;
					}

					case 'V':
					{
						VELOCITY_DELTA = atof( Argv[k++] );

						break;
					}

					case 'A':
					{
						ANGLE_DELTA = atof( Argv[k++] );

						break;
					}

					case 'S':
					{
						SPEED_DELTA = atof( Argv[k++] );

						break;
					}

					case 'W':
					{
						WANDER_PROB = atof( Argv[k++] );

						break;
					}

					case 'F':
					{
						FIRE_PROB = atof( Argv[k++] );

						break;
					}

					case 'C':
					{
						HELI_PROB = atof( Argv[k++] );

						break;
					}

					case 'R':
					{
						HELI_RADIUS = atoi( Argv[k++] );

						break;
					}

					case 'v':
					{
						vflag++;

						stv_setopt( stvOptNull, stvOptVerbose,
							vflag ? STV_DEBUG_ALL : 0 );

						break;
					}

					case 'd': dflag++; break;

					case 'c': cflag++; break;

					case 'h': do_usage++; break;

					default:
					{
						printf( "Unknown Option -%c\n", Argv[i][j] );

						do_usage++;

						break;
					}
				}
			}
		}

		i = k - 1;
	}

	if ( do_usage )
		usage();
}


void
usage()
{
	printf( "\nusage:  flyme [-vd] [-N <ntasks>] [-SPMD]\n" );
	printf( "\t[-B <speed>] [-[V|A|S] <delta>] [-[W|F|H] <prob>]\n" );
	printf( "\t[-WA <wind_angle>] [-WM <wind_magn>]\n" );

	printf( "\nwhere:\n" );
	printf( "------\n" );

	printf( "\t-v = Verbose Operation (%s)\n", vflag ? "On" : "Off" );

	printf( "\t-d = Spawn in Debug Mode (%s)\n", dflag ? "On" : "Off" );

	printf( "\t-c = Verbose for Collision Detect (%s)\n",
		cflag ? "On" : "Off" );

	printf( "\t-N <ntasks> = # of Initial Processors (%d)\n", NTASKS );

	printf( "\t-SPMD = do not use master to spawn rest of tasks (%s)\n",
		SPMD ? "On" : "Off" );

	printf(
	"\t-H <speed> = Set Speed of Helicopters (%lf) (Bullets X %lf)\n",
		HELI_VELOCITY, (double) SPEED_FACTOR );

	printf( "\t-B <speed> = Set Speed of Bullets (%lf)\n",
		BULLET_VELOCITY );

	printf( "\t-V <delta> = Set Velocity Angle Change Delta (%lf)\n",
		VELOCITY_DELTA );

	printf( "\t-A <delta> = Set Heading Angle Change Delta (%lf)\n",
		ANGLE_DELTA );

	printf( "\t-S <delta> = Set Speed Change Delta (%lf)\n",
		SPEED_DELTA );

	printf( "\t-W <prob> = Set Wandering Probability (%lf)\n",
		WANDER_PROB );

	printf( "\t-F <prob> = Set Firing Probability (%lf)\n", FIRE_PROB );

	printf( "\t-C <prob> = Set Helicopter Probability (%lf)\n",
		HELI_PROB );

	printf( "\t-R <prob> = Set Helicopter Radius (%d)\n",
		HELI_RADIUS );

	printf( "\t-WA <wind_angle> = Set Wind Angle (%lf)\n", WIND_ANGLE );

	printf( "\t-WM <wind_magn> = Set Wind Speed (%lf)\n", WIND_MAGN );

	printf( "\n" );

	exit( 0 );
}


void
program_init()
{
	/* Seed Random Generator */

	srandom( getpid() * (int) time( (time_t *) NULL ) );

	/* Initialize Globals */

	GLOBREGION = (REGION) NULL;

	MYREGION = (REGION) NULL;

	HELIS = (HELI) NULL;

	BULLETS = (BULLET) NULL;

	EXPLOSIONS = (EXPLOSION) NULL;

	BLIPS = (BLIP) NULL;

	NEXT_EXPLOSION_ID = 0;

	NEXT_BULLET_ID = 0;

	NEXT_HELI_ID = 0;

	/* Steering Parameters */

	HELI_HEAD_GOAL = HELI_VEL_GOAL = HELI_RADAR_RANGE = -1.0;

	HELI_SHIELD = HELI_FIRE = HELI_INDEX = HELI_RELEASE = -1;
}


void
pvm_init()
{
	int flags;
	int left;
	int cnt;
	int num;

	/* Join PVM */

	MYTID = pvm_mytid();

	printf( "My TID = [ 0x%x / %d ]\n", MYTID, MYTID );
	fflush( stdout );

	/* Join Application Group */

	MYINST = pvm_joingroup( GROUPNAME );

	if ( MYINST < 0 )
	{
		pvm_perror( "Joining Group" );

		exit( 0 );
	}

	/* Determine if Master - Spawn Tasks if Necessary */

	if ( !SPMD )
	{
		if ( !MYINST && NTASKS > 1 )
		{
			printf( "Spawning %d Other Tasks... ", NTASKS - 1 );
			fflush( stdout );

			if ( dflag )
				flags = PvmTaskDefault | PvmTaskDebug;

			else
				flags = PvmTaskDefault;

			left = NTASKS - 1;

			num = 0;

			cnt = 0;

			do
			{
				left -= num;

				num = pvm_spawn( "region", Argv + 1, flags,
					(char *) NULL, left, (int *) NULL );

				if ( num < 0 )
				{
					printf( "\n\nPVM System Error - Exiting.\n\n" );
					fflush( stdout );

					exit( 0 );
				}

				else if ( num != left )
				{
					printf( "\n\tOnly Spawned %d Tasks, Retrying... ",
						num );
					fflush( stdout );
				}

				cnt++;
			}
			while ( num != left && cnt < 3 );

			if ( num != left )
			{
				printf( "\n\nGiving Up on Spawn.  Exiting.\n\n" );
				printf( "Please Reset the Virtual Machine.\n\n" );
				fflush( stdout );

				exit( 0 );
			}

			else
			{
				printf( "done.\n" );
				fflush( stdout );
			}
		}
	}
}


void
region_init()
{
	REGION R;

	int tids[16];

	int neigh;
	int slice;
	int i;

	/* Free Up Existing Regions */

	if ( GLOBREGION != NULL )
		free_region( &GLOBREGION );

	if ( MYREGION != NULL )
		free_region_list( &MYREGION );

	/* Define Global Region */

	GLOBREGION = create_region();

	GLOBREGION->xmin = 0;
	GLOBREGION->xmax = FIELD_WIDTH - 1;

	GLOBREGION->ymin = 0;
	GLOBREGION->ymax = FIELD_HEIGHT - 1;

	/* Slice Size */

	slice = FIELD_WIDTH / NTASKS;

	/* My SLice */

	MYREGION = create_region();

	MYREGION->tid = MYTID;

	MYREGION->xmin = MYINST * slice;
	MYREGION->xmax = MYREGION->xmin + slice - 1;

	MYREGION->ymin = GLOBREGION->ymin;
	MYREGION->ymax = GLOBREGION->ymax;

	MYREGION->nneigh = 0;

	R = MYREGION;

	/* Left Neighbor */

	neigh = MYINST - 1;

	if ( neigh >= 0 )
	{
		R = R->neigh = create_region();

		R->xmin = neigh * slice;
		R->xmax = R->xmin + slice - 1;

		R->ymin = GLOBREGION->ymin;
		R->ymax = GLOBREGION->ymax;

		while ( (R->tid = pvm_gettid( GROUPNAME, neigh )) < 0 );

		(MYREGION->nneigh)++;
	}

	/* Right Neighbor */

	neigh = MYINST + 1;

	if ( neigh < NTASKS )
	{
		R = R->neigh = create_region();

		R->xmin = neigh * slice;
		R->xmax = R->xmin + slice - 1;

		R->ymin = GLOBREGION->ymin;
		R->ymax = GLOBREGION->ymax;

		while ( (R->tid = pvm_gettid( GROUPNAME, neigh )) < 0 );

		(MYREGION->nneigh)++;
	}

	/* Set Up Notifies */

	R = MYREGION->neigh;

	i = 0;

	while ( R != NULL )
	{
		tids[ i++ ] = R->tid;

		R = R->neigh;
	}

	if ( i )
		pvm_notify( PvmTaskExit, TRANSFER_MSG, i, tids );
}


void
heli_init()
{
	/* Everybody Gets One Helicopter to Start With */

	add_rand_heli();
}


void
do_stv_stuff()
{
	STV_VALUE heli_params[7];
	STV_VALUE wind_params[2];

	STV_VALUE vptr;

	char *heli_names[7];
	char *wind_names[2];

	char *sptr;

	int heli_types[7];
	int wind_types[2];

	int glb[2];
	int gub[2];

	int hpart_id;
	int bpart_id;
	int epart_id;

	int type;

	stv_init( APPNAME, STV_MSGTAG, NTASKS, MYINST );

	region_init();

	glb[0] = GLOBREGION->xmin;
	glb[1] = GLOBREGION->ymin;

	gub[0] = GLOBREGION->xmax;
	gub[1] = GLOBREGION->ymax;

	/* Helicopter Particle / Fields */

	hpart_id = stv_particleDefine( "helicopter", 2, glb, gub, NTASKS,
		get_helicopter, make_helicopter );

	stv_pfieldDefine( "hid", hpart_id, stvInt, 1,
		get_helicopter_pfield, (STV_VALUE) PF_HELI_HID,
		set_helicopter_pfield, (STV_VALUE) PF_HELI_HID,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "fire_count", hpart_id, stvInt, 1,
		get_helicopter_pfield, (STV_VALUE) PF_HELI_FIRE_COUNT,
		set_helicopter_pfield, (STV_VALUE) PF_HELI_FIRE_COUNT,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "steered", hpart_id, stvInt, 1,
		get_helicopter_pfield, (STV_VALUE) PF_HELI_STEERED,
		set_helicopter_pfield, (STV_VALUE) PF_HELI_STEERED,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "shield_str", hpart_id, stvInt, 1,
		get_helicopter_pfield, (STV_VALUE) PF_HELI_SHIELD_STR,
		set_helicopter_pfield, (STV_VALUE) PF_HELI_SHIELD_STR,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "shielded", hpart_id, stvInt, 1,
		get_helicopter_pfield, (STV_VALUE) PF_HELI_SHIELDED,
		set_helicopter_pfield, (STV_VALUE) PF_HELI_SHIELDED,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "score", hpart_id, stvInt, 1,
		get_helicopter_pfield, (STV_VALUE) PF_HELI_SCORE,
		set_helicopter_pfield, (STV_VALUE) PF_HELI_SCORE,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "radar_range", hpart_id, stvDouble, 1,
		get_helicopter_pfield, (STV_VALUE) PF_HELI_RADAR_RANGE,
		set_helicopter_pfield, (STV_VALUE) PF_HELI_RADAR_RANGE,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "head_goal", hpart_id, stvDouble, 1,
		get_helicopter_pfield, (STV_VALUE) PF_HELI_HEAD_GOAL,
		set_helicopter_pfield, (STV_VALUE) PF_HELI_HEAD_GOAL,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "heading", hpart_id, stvDouble, 1,
		get_helicopter_pfield, (STV_VALUE) PF_HELI_HEADING,
		set_helicopter_pfield, (STV_VALUE) PF_HELI_HEADING,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "hvel_angle", hpart_id, stvDouble, 1,
		get_helicopter_pfield, (STV_VALUE) PF_HELI_VEL_ANGLE,
		set_helicopter_pfield, (STV_VALUE) PF_HELI_VEL_ANGLE,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "hvel_magn", hpart_id, stvDouble, 1,
		get_helicopter_pfield, (STV_VALUE) PF_HELI_VEL_MAGN,
		set_helicopter_pfield, (STV_VALUE) PF_HELI_VEL_MAGN,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "hvel_goal", hpart_id, stvDouble, 1,
		get_helicopter_pfield, (STV_VALUE) PF_HELI_VEL_GOAL,
		set_helicopter_pfield, (STV_VALUE) PF_HELI_VEL_GOAL,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "wimp_factor", hpart_id, stvDouble, 1,
		get_helicopter_pfield, (STV_VALUE) PF_HELI_WIMP_FACTOR,
		set_helicopter_pfield, (STV_VALUE) PF_HELI_WIMP_FACTOR,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "last_bdist", hpart_id, stvDouble, 1,
		get_helicopter_pfield, (STV_VALUE) PF_HELI_LAST_BDIST,
		set_helicopter_pfield, (STV_VALUE) PF_HELI_LAST_BDIST,
		stvVisOnly | stvCpOnly );

	/* Bullet Particle / Fields */

	bpart_id = stv_particleDefine( "bullet", 2, glb, gub, NTASKS,
		get_bullet, make_bullet );

	stv_pfieldDefine( "bid", bpart_id, stvInt, 1,
		get_bullet_pfield, (STV_VALUE) PF_BULLET_BID,
		set_bullet_pfield, (STV_VALUE) PF_BULLET_BID,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "age", bpart_id, stvInt, 1,
		get_bullet_pfield, (STV_VALUE) PF_BULLET_AGE,
		set_bullet_pfield, (STV_VALUE) PF_BULLET_AGE,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "bvel_angle", bpart_id, stvDouble, 1,
		get_bullet_pfield, (STV_VALUE) PF_BULLET_VEL_ANGLE,
		set_bullet_pfield, (STV_VALUE) PF_BULLET_VEL_ANGLE,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "bvel_magn", bpart_id, stvDouble, 1,
		get_bullet_pfield, (STV_VALUE) PF_BULLET_VEL_MAGN,
		set_bullet_pfield, (STV_VALUE) PF_BULLET_VEL_MAGN,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "parent", bpart_id, stvInt, 1,
		get_bullet_pfield, (STV_VALUE) PF_BULLET_PARENT,
		set_bullet_pfield, (STV_VALUE) PF_BULLET_PARENT,
		stvVisOnly | stvCpOnly );

	/* Explosion Particle / Fields */

	epart_id = stv_particleDefine( "explosion", 2, glb, gub, NTASKS,
		get_explosion, make_explosion );

	stv_pfieldDefine( "eid", epart_id, stvInt, 1,
		get_explosion_pfield, (STV_VALUE) PF_EXPLOSION_EID,
		set_explosion_pfield, (STV_VALUE) PF_EXPLOSION_EID,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "explosion_state", epart_id, stvInt, 1,
		get_explosion_pfield, (STV_VALUE) PF_EXPLOSION_STATE,
		set_explosion_pfield, (STV_VALUE) PF_EXPLOSION_STATE,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "explosion_angle", epart_id, stvDouble, 1,
		get_explosion_pfield, (STV_VALUE) PF_EXPLOSION_ANGLE,
		set_explosion_pfield, (STV_VALUE) PF_EXPLOSION_ANGLE,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "explosion_strength", epart_id, stvInt, 1,
		get_explosion_pfield, (STV_VALUE) PF_EXPLOSION_STRENGTH,
		set_explosion_pfield, (STV_VALUE) PF_EXPLOSION_STRENGTH,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "explosion_radius", epart_id, stvInt, 1,
		get_explosion_pfield, (STV_VALUE) PF_EXPLOSION_RADIUS,
		set_explosion_pfield, (STV_VALUE) PF_EXPLOSION_RADIUS,
		stvVisOnly | stvCpOnly );

	/* Steering Parameters */

	/* Wind */

	wind_params[0] = (STV_VALUE) &WIND_ANGLE;
	wind_params[1] = (STV_VALUE) &WIND_MAGN;

	wind_names[0] = "angle";
	wind_names[1] = "magnitude";

	wind_types[0] = stvDouble;
	wind_types[1] = stvDouble;

	stv_vparamDefine( "wind", wind_params, wind_names, wind_types, 2,
		stvVisOnly | stvCpOnly );

	/* Helis */

	heli_params[0] = (STV_VALUE) &HELI_HEAD_GOAL;
	heli_params[1] = (STV_VALUE) &HELI_VEL_GOAL;
	heli_params[2] = (STV_VALUE) &HELI_RADAR_RANGE;
	heli_params[3] = (STV_VALUE) &HELI_FIRE;
	heli_params[4] = (STV_VALUE) &HELI_SHIELD;
	heli_params[5] = (STV_VALUE) &HELI_INDEX;
	heli_params[6] = (STV_VALUE) &HELI_RELEASE;

	heli_names[0] = "heading goal";
	heli_names[1] = "velocity goal";
	heli_names[2] = "radar range";
	heli_names[3] = "fire bullet";
	heli_names[4] = "shielded";
	heli_names[5] = "index";
	heli_names[6] = "release control";

	heli_types[0] = stvDouble;
	heli_types[1] = stvDouble;
	heli_types[2] = stvDouble;
	heli_types[3] = stvInt;
	heli_types[4] = stvInt;
	heli_types[5] = stvIndex | stvInt;
	heli_types[6] = stvInt;

	stv_vparamDefine( "helicopter",
		heli_params, heli_names, heli_types, 7,
		stvVisOnly | stvCpOnly );

	/* Other Parameters */

	sptr = "heli velocity";
	vptr = (STV_VALUE) &HELI_VELOCITY;
	type = stvDouble;

	stv_paramDefine( sptr, vptr, type, stvVisOnly | stvCpOnly );

	sptr = "bullet velocity";
	vptr = (STV_VALUE) &BULLET_VELOCITY;
	type = stvDouble;

	stv_paramDefine( sptr, vptr, type, stvVisOnly | stvCpOnly );

	sptr = "velocity delta";
	vptr = (STV_VALUE) &VELOCITY_DELTA;
	type = stvDouble;

	stv_paramDefine( sptr, vptr, type, stvVisOnly | stvCpOnly );

	sptr = "angle delta";
	vptr = (STV_VALUE) &ANGLE_DELTA;
	type = stvDouble;

	stv_paramDefine( sptr, vptr, type, stvVisOnly | stvCpOnly );

	sptr = "speed delta";
	vptr = (STV_VALUE) &SPEED_DELTA;
	type = stvDouble;

	stv_paramDefine( sptr, vptr, type, stvVisOnly | stvCpOnly );

	sptr = "wander prob";
	vptr = (STV_VALUE) &WANDER_PROB;
	type = stvDouble;

	stv_paramDefine( sptr, vptr, type, stvVisOnly | stvCpOnly );

	sptr = "fire prob";
	vptr = (STV_VALUE) &FIRE_PROB;
	type = stvDouble;

	stv_paramDefine( sptr, vptr, type, stvVisOnly | stvCpOnly );

	sptr = "heli prob";
	vptr = (STV_VALUE) &HELI_PROB;
	type = stvDouble;

	stv_paramDefine( sptr, vptr, type, stvVisOnly | stvCpOnly );

	sptr = "heli radius";
	vptr = (STV_VALUE) &HELI_RADIUS;
	type = stvInt;

	stv_paramDefine( sptr, vptr, type, stvVisOnly | stvCpOnly );

	/* Checkpointed Only Parameters */

	sptr = "next explosion id";
	vptr = (STV_VALUE) &NEXT_EXPLOSION_ID;
	type = stvInt;

	stv_paramDefine( sptr, vptr, type, stvCpOnly );

	sptr = "next bullet id";
	vptr = (STV_VALUE) &NEXT_BULLET_ID;
	type = stvInt;

	stv_paramDefine( sptr, vptr, type, stvCpOnly );

	sptr = "next heli id";
	vptr = (STV_VALUE) &NEXT_HELI_ID;
	type = stvInt;

	stv_paramDefine( sptr, vptr, type, stvCpOnly );

	sptr = "vflag";
	vptr = (STV_VALUE) &vflag;
	type = stvInt;

	stv_paramDefine( sptr, vptr, type, stvCpOnly );

	sptr = "dflag";
	vptr = (STV_VALUE) &dflag;
	type = stvInt;

	stv_paramDefine( sptr, vptr, type, stvCpOnly );

	sptr = "cflag";
	vptr = (STV_VALUE) &cflag;
	type = stvInt;

	stv_paramDefine( sptr, vptr, type, stvCpOnly );
}


REGION create_region()
{
	REGION tmp;

	tmp = (REGION) malloc( sizeof( struct region_struct ) );
	memcheck( tmp, "Region Structure" );

	tmp->tid = -1;

	tmp->xmin = -1;
	tmp->xmax = -1;

	tmp->ymin = -1;
	tmp->ymax = -1;

	tmp->nneigh = -1;

	tmp->neigh = (REGION) NULL;

	return( tmp );
}


void
free_region( ptr )
REGION *ptr;
{
	REGION R;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	R = *ptr;

	R->tid = -1;

	R->xmin = -1;
	R->xmax = -1;

	R->ymin = -1;
	R->ymax = -1;

	R->nneigh = -1;

	R->neigh = (REGION) NULL;

	free( R );

	*ptr = (REGION) NULL;
}


void
free_region_list( ptr )
REGION *ptr;
{
	REGION Rneigh;
	REGION R;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	R = *ptr;

	*ptr = (REGION) NULL;

	while ( R != NULL )
	{
		Rneigh = R->neigh;

		free_region( &R );

		R = Rneigh;
	}
}


HELI create_heli()
{
	HELI tmp;

	tmp = (HELI) malloc( sizeof( struct heli_struct ) );
	memcheck( tmp, "Helicopter Structure" );

	tmp->hid = -1;

	tmp->loc_x = -1;
	tmp->loc_y = -1;

	tmp->fire_count = -1;

	tmp->killed = -1;

	tmp->steered = -1;

	tmp->shield_str = -1;

	tmp->shielded = -1;

	tmp->score = -1;

	tmp->bounce = -1;

	tmp->radar_range = -1.0;

	tmp->head_goal = -1.0;
	tmp->heading = -1.0;

	tmp->vel_angle = -1.0;
	tmp->vel_magn = -1.0;
	tmp->vel_goal = -1.0;

	tmp->wimp_factor = -1.0;
	tmp->last_bdist = -1.0;

	tmp->next = (HELI) NULL;

	return( tmp );
}


void
free_heli( ptr )
HELI *ptr;
{
	HELI H;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	H = *ptr;

	H->hid = -1;

	H->loc_x = -1;
	H->loc_y = -1;

	H->fire_count = -1;

	H->killed = -1;

	H->steered = -1;

	H->shield_str = -1;

	H->shielded = -1;

	H->score = -1;

	H->bounce = -1;

	H->radar_range = -1.0;

	H->head_goal = -1.0;
	H->heading = -1.0;

	H->vel_angle = -1.0;
	H->vel_magn = -1.0;
	H->vel_goal = -1.0;

	H->wimp_factor = -1.0;
	H->last_bdist = -1.0;

	H->next = (HELI) NULL;

	free( H );

	*ptr = (HELI) NULL;
}


HELI add_heli( hid, x, y, fire_count, steered, shield_str, shielded,
	score, radar_range, head_goal, heading, vel_angle, vel_magn,
	vel_goal, wimp_factor, last_bdist )
int hid;
int x, y;
int fire_count;
int steered;
int shield_str;
int shielded;
int score;
double radar_range;
double head_goal;
double heading;
double vel_angle;
double vel_magn;
double vel_goal;
double wimp_factor;
double last_bdist;
{
	HELI H;

	if ( HELIS == NULL )
		H = HELIS = create_heli();
	
	else
	{
		H = HELIS;

		while ( H->next != NULL )
			H = H->next;

		H = H->next = create_heli();
	}

	DUMP_HELI( "adding", hid, x, y, radar_range, heading, head_goal,
		vel_angle, vel_magn, vel_goal );

	H->hid = hid;

	H->loc_x = x;
	H->loc_y = y;

	H->fire_count = fire_count;

	H->killed = stvFalse;

	H->steered = steered;

	H->shield_str = shield_str;

	H->shielded = shielded;

	H->score = score;

	H->radar_range = radar_range;

	H->head_goal = head_goal;
	H->heading = heading;

	H->vel_angle = vel_angle;

	H->vel_magn = vel_magn;
	H->vel_goal = vel_goal;

	H->wimp_factor = wimp_factor;
	H->last_bdist = last_bdist;

	H->bounce = stvFalse;

	return( H );
}


void
add_rand_heli()
{
	double radar_range;

	double head_goal;
	double heading;

	double vel_angle;
	double vel_magn;
	double vel_goal;

	double wimp_factor;

	int x, y;
	int hid;

	/* Generate One Random Helicopter */

	if ( vflag )
	{
		printf( "\nadding random helicopter:\n" );
		fflush( stdout );
	}

	printf( "H" ); fflush( stdout );

	hid = HELI_ID( MYINST );

	rand_point( MYREGION, &x, &y );

	radar_range = rand_double() * (double) MAX_RADAR_RANGE;

	head_goal = rand_angle();
	heading = rand_angle();

	vel_angle = rand_angle();

	vel_magn = rand_double() * HELI_VELOCITY;
	vel_goal = rand_double() * HELI_VELOCITY;

	wimp_factor = rand_double();

	add_heli( hid, x, y, 0, stvFalse, SHIELD_STRENGTH, stvFalse, 0,
		radar_range, head_goal, heading, vel_angle, vel_magn, vel_goal,
		wimp_factor, -1.0 );
}


HELI get_heli_by_hid( hid )
int hid;
{
	HELI H;

	H = HELIS;

	while ( H != NULL )
	{
		if ( H->hid == hid )
			return( H );
		
		H = H->next;
	}

	return( (HELI) NULL );
}


void
fire_bullet( H, str )
HELI H;
char *str;
{
	int bid;

	if ( H->fire_count == 0 )
	{
		if ( vflag )
		{
			printf( "\n*** Helicopter %d Fires!!! (%s) ***\n",
				H->hid, str );
			fflush( stdout );
		}

		bid = BULLET_ID( MYINST );

		add_bullet( bid, 0, H->loc_x, H->loc_y, H->heading,
			BULLET_VELOCITY, H->hid );

		H->fire_count += FIRE_DELAY;
	}
}


damage_heli( H, hx, hy, strength )
HELI H;
double hx, hy;
int strength;
{
	if ( vflag )
	{
		printf( "*** Damage Heli %d at (%lf,%lf)\n", H->hid, hx, hy );
		fflush( stdout );
	}

	if ( H->shielded )
	{
		H->shield_str -= strength;

		if ( H->shield_str < 0 )
			H->killed = stvTrue;
	}

	else
		H->killed = stvTrue;

	return( H->killed );
}


BULLET create_bullet()
{
	BULLET tmp;

	tmp = (BULLET) malloc( sizeof( struct bullet_struct ) );
	memcheck( tmp, "Bullet Structure" );

	tmp->bid = -1;

	tmp->age = -1;

	tmp->loc_x = -1;
	tmp->loc_y = -1;

	tmp->vel_angle = -1.0;
	tmp->vel_magn = -1.0;

	tmp->parent = -1;

	tmp->next = (BULLET) NULL;

	return( tmp );
}


void
free_bullet( ptr )
BULLET *ptr;
{
	BULLET B;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	B = *ptr;

	B->bid = -1;

	B->age = -1;

	B->loc_x = -1;
	B->loc_y = -1;

	B->vel_angle = -1.0;
	B->vel_magn = -1.0;

	B->parent = -1;

	B->next = (BULLET) NULL;

	free( B );

	*ptr = (BULLET) NULL;
}


BULLET add_bullet( bid, age, x, y, heading, velocity, parent )
int bid;
int age;
int x, y;
double heading;
double velocity;
int parent;
{
	BULLET B;

	if ( BULLETS == NULL )
		B = BULLETS = create_bullet();
	
	else
	{
		B = BULLETS;

		while ( B->next != NULL )
			B = B->next;

		B = B->next = create_bullet();
	}

	DUMP_BULLET( "adding", bid, x, y, heading, velocity );

	B->bid = bid;

	B->age = age;

	B->loc_x = x;
	B->loc_y = y;

	B->vel_angle = heading;
	B->vel_magn = velocity;

	B->parent = parent;

	return( B );
}


EXPLOSION create_explosion()
{
	EXPLOSION tmp;

	tmp = (EXPLOSION) malloc( sizeof( struct explosion_struct ) );
	memcheck( tmp, "Explosion Structure" );

	tmp->eid = -1;

	tmp->state = -1;

	tmp->loc_x = -1;
	tmp->loc_y = -1;

	tmp->angle = -1.0;

	tmp->strength = -1;

	tmp->radius = -1;

	tmp->next = (EXPLOSION) NULL;

	return( tmp );
}


void
free_explosion( ptr )
EXPLOSION *ptr;
{
	EXPLOSION E;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	E = *ptr;

	E->eid = -1;

	E->state = -1;

	E->loc_x = -1;
	E->loc_y = -1;

	E->angle = -1.0;

	E->strength = -1;

	E->radius = -1;

	E->next = (EXPLOSION) NULL;

	free( E );

	*ptr = (EXPLOSION) NULL;
}


EXPLOSION add_explosion( x, y, angle, strength )
int x, y;
double angle;
int strength;
{
	EXPLOSION E;

	int eid;

	if ( EXPLOSIONS == NULL )
		E = EXPLOSIONS = create_explosion();
	
	else
	{
		E = EXPLOSIONS;

		while ( E->next != NULL )
			E = E->next;

		E = E->next = create_explosion();
	}

	eid = EXPLOSION_ID( MYINST );

	DUMP_EXPLOSION( "adding", eid, x, y, angle, strength );

	E->eid = eid;

	E->state = EXPLOSION_GROWING;

	E->loc_x = x;
	E->loc_y = y;

	E->angle = angle;

	E->strength = strength;

	E->radius = 0;

	return( E );
}


BLIP create_blip()
{
	BLIP tmp;

	tmp = (BLIP) malloc( sizeof( struct blip_struct ) );
	memcheck( tmp, "Blip Structure" );

	tmp->loc_x = -1;
	tmp->loc_y = -1;

	tmp->radius = -1.0;

	tmp->next = (BLIP) NULL;

	return( tmp );
}


void
free_blip( ptr )
BLIP *ptr;
{
	BLIP BP;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	BP = *ptr;

	BP->loc_x = -1;
	BP->loc_y = -1;

	BP->radius = -1.0;

	BP->next = (BLIP) NULL;

	free( BP );

	*ptr = (BLIP) NULL;
}


void
free_blip_list( ptr )
BLIP *ptr;
{
	BLIP BPnext;
	BLIP BP;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	BP = *ptr;

	*ptr = (BLIP) NULL;

	while ( BP != NULL )
	{
		BPnext = BP->next;

		free_blip( &BP );

		BP = BPnext;
	}
}


BLIP add_blip( x, y, radius )
int x, y;
double radius;
{
	BLIP BP;

	if ( BLIPS == NULL )
		BP = BLIPS = create_blip();
	
	else
	{
		BP = BLIPS;

		while ( BP->next != NULL )
			BP = BP->next;

		BP = BP->next = create_blip();
	}

	BP->loc_x = x;
	BP->loc_y = y;

	BP->radius = radius;

	return( BP );
}


void
cleanup_region()
{
	EXPLOSION Enext;
	EXPLOSION E;

	BULLET Bnext;
	BULLET B;

	BLIP BPnext;
	BLIP BP;

	HELI Hnext;
	HELI H;

	/* Free Any Helicopters */

	H = HELIS;

	HELIS = (HELI) NULL;

	while ( H != NULL )
	{
		Hnext = H->next;

		free_heli( &H );

		H = Hnext;
	}

	/* Free Any Bullets */

	B = BULLETS;

	BULLETS = (BULLET) NULL;

	while ( B != NULL )
	{
		Bnext = B->next;

		free_bullet( &B );

		B = Bnext;
	}

	/* Free Any Exlposions */

	E = EXPLOSIONS;

	EXPLOSIONS = (EXPLOSION) NULL;

	while ( E != NULL )
	{
		Enext = E->next;

		free_explosion( &E );

		E = Enext;
	}

	/* Free Any Blips */

	BP = BLIPS;

	BLIPS = (BLIP) NULL;

	while ( BP != NULL )
	{
		BPnext = BP->next;

		free_blip( &BP );

		BP = BPnext;
	}
}


void
reset_steered()
{
	HELI H;

	H = HELIS;

	while ( H != NULL )
	{
		H->steered = stvFalse;

		H = H->next;
	}
}


void
get_helicopter( hnum, R, id, coords )
int hnum;
STV_REGION R;
STV_PARTICLE_ID *id;
int *coords;
{
	static HELI H_save = (HELI) NULL;
	static int count_save = -1;

	HELI H;

	int hcoords[2];

	int count;

	/* Reset */

	if ( hnum == -1 )
	{
		*id = (STV_PARTICLE_ID) NULL;

		H_save = (HELI) NULL;

		count_save = -1;
	}

	/* Restore */

	if ( hnum >= count_save && count_save != -1 )
	{
		H = H_save;

		count = count_save;
	}

	/* Fresh Search */

	else
	{
		H = HELIS;

		count = 0;
	}

	/* Search for Helicopter */

	while ( H != NULL )
	{
		hcoords[0] = H->loc_x;
		hcoords[1] = H->loc_y;

		if ( stv_particle_in_region( R, hcoords, 2 ) )
		{
			count++;

			if ( count == hnum )
			{
				*id = (STV_PARTICLE_ID) H;

				coords[0] = hcoords[0];
				coords[1] = hcoords[1];

				H_save = H->next;

				count_save = count;

				return;
			}
		}

		H = H->next;
	}

	/* Helicopter Not Found */

	*id = (STV_PARTICLE_ID) NULL;

	H_save = (HELI) NULL;

	count_save = -1;
}


void
get_helicopter_pfield( id, field, value )
STV_PARTICLE_ID id;
STV_VALUE field;
STV_VALUE *value;
{
	HELI H;

	long fid;

	H = (HELI) id;

	fid = (long) field;

	switch ( fid )
	{
		case PF_HELI_HID:
			*value = (STV_VALUE) &(H->hid);
			break;

		case PF_HELI_FIRE_COUNT:
			*value = (STV_VALUE) &(H->fire_count);
			break;

		case PF_HELI_STEERED:
			*value = (STV_VALUE) &(H->steered);
			break;

		case PF_HELI_SHIELD_STR:
			*value = (STV_VALUE) &(H->shield_str);
			break;

		case PF_HELI_SHIELDED:
			*value = (STV_VALUE) &(H->shielded);
			break;

		case PF_HELI_SCORE:
			*value = (STV_VALUE) &(H->score);
			break;

		case PF_HELI_RADAR_RANGE:
			*value = (STV_VALUE) &(H->radar_range);
			break;

		case PF_HELI_HEAD_GOAL:
			*value = (STV_VALUE) &(H->head_goal);
			break;

		case PF_HELI_HEADING:
			*value = (STV_VALUE) &(H->heading);
			break;

		case PF_HELI_VEL_ANGLE:
			*value = (STV_VALUE) &(H->vel_angle);
			break;

		case PF_HELI_VEL_MAGN:
			*value = (STV_VALUE) &(H->vel_magn);
			break;

		case PF_HELI_VEL_GOAL:
			*value = (STV_VALUE) &(H->vel_goal);
			break;

		case PF_HELI_WIMP_FACTOR:
			*value = (STV_VALUE) &(H->wimp_factor);
			break;

		case PF_HELI_LAST_BDIST:
			*value = (STV_VALUE) &(H->last_bdist);
			break;

		default:
			*value = (STV_VALUE) NULL;
			break;
	}
}


void
make_helicopter( coords, id )
int *coords;
STV_PARTICLE_ID *id;
{
	HELI H;

	if ( HELIS == NULL )
		H = HELIS = create_heli();
	
	else
	{
		H = HELIS;

		while ( H->next != NULL )
			H = H->next;

		H = H->next = create_heli();
	}

	H->loc_x = coords[0];
	H->loc_y = coords[1];

	H->killed = stvFalse;

	H->bounce = stvFalse;

	*id = (STV_PARTICLE_ID) H;
}


void
set_helicopter_pfield( id, field, value )
STV_PARTICLE_ID id;
STV_VALUE field;
STV_VALUE value;
{
	HELI H;

	long fid;

	H = (HELI) id;

	fid = (long) field;

	switch ( fid )
	{
		case PF_HELI_HID:
			H->hid = STV_VALUE_OF( value, int );
			break;

		case PF_HELI_FIRE_COUNT:
			H->fire_count = STV_VALUE_OF( value, int );
			break;

		case PF_HELI_STEERED:
			H->steered = STV_VALUE_OF( value, int );
			break;

		case PF_HELI_SHIELD_STR:
			H->shield_str = STV_VALUE_OF( value, int );
			break;

		case PF_HELI_SHIELDED:
			H->shielded = STV_VALUE_OF( value, int );
			break;

		case PF_HELI_SCORE:
			H->score = STV_VALUE_OF( value, int );
			break;

		case PF_HELI_RADAR_RANGE:
			H->radar_range = STV_VALUE_OF( value, double );
			break;

		case PF_HELI_HEAD_GOAL:
			H->head_goal = STV_VALUE_OF( value, double );
			break;

		case PF_HELI_HEADING:
			H->heading = STV_VALUE_OF( value, double );
			break;

		case PF_HELI_VEL_ANGLE:
			H->vel_angle = STV_VALUE_OF( value, double );
			break;

		case PF_HELI_VEL_MAGN:
			H->vel_magn = STV_VALUE_OF( value, double );
			break;

		case PF_HELI_VEL_GOAL:
			H->vel_goal = STV_VALUE_OF( value, double );
			break;

		case PF_HELI_WIMP_FACTOR:
			H->wimp_factor = STV_VALUE_OF( value, double );
			break;

		case PF_HELI_LAST_BDIST:
			H->last_bdist = STV_VALUE_OF( value, double );
			break;

		default:
			break;
	}
}


void
get_bullet( bnum, R, id, coords )
int bnum;
STV_REGION R;
STV_PARTICLE_ID *id;
int *coords;
{
	static BULLET B_save = (BULLET) NULL;
	static int count_save = -1;

	BULLET B;

	int bcoords[2];

	int count;

	/* Reset */

	if ( bnum == -1 )
	{
		*id = (STV_PARTICLE_ID) NULL;

		B_save = (BULLET) NULL;

		count_save = -1;
	}

	/* Restore */

	if ( bnum >= count_save && count_save != -1 )
	{
		B = B_save;

		count = count_save;
	}

	/* Fresh Search */

	else
	{
		B = BULLETS;

		count = 0;
	}

	/* Search for Bullet */

	while ( B != NULL )
	{
		bcoords[0] = B->loc_x;
		bcoords[1] = B->loc_y;

		if ( stv_particle_in_region( R, bcoords, 2 ) )
		{
			count++;

			if ( count == bnum )
			{
				*id = (STV_PARTICLE_ID) B;

				coords[0] = bcoords[0];
				coords[1] = bcoords[1];

				B_save = B->next;

				count_save = count;

				return;
			}
		}

		B = B->next;
	}

	/* Bullet Not Found */

	*id = (STV_PARTICLE_ID) NULL;

	B_save = (BULLET) NULL;

	count_save = -1;
}


void
get_bullet_pfield( id, field, value )
STV_PARTICLE_ID id;
STV_VALUE field;
STV_VALUE *value;
{
	BULLET B;

	long fid;

	B = (BULLET) id;

	fid = (long) field;

	switch ( fid )
	{
		case PF_BULLET_BID:
			*value = (STV_VALUE) &(B->bid);
			break;

		case PF_BULLET_AGE:
			*value = (STV_VALUE) &(B->age);
			break;

		case PF_BULLET_VEL_ANGLE:
			*value = (STV_VALUE) &(B->vel_angle);
			break;

		case PF_BULLET_VEL_MAGN:
			*value = (STV_VALUE) &(B->vel_magn);
			break;

		case PF_BULLET_PARENT:
			*value = (STV_VALUE) &(B->parent);
			break;

		default:
			*value = (STV_VALUE) NULL;
			break;
	}
}


void
make_bullet( coords, id )
int *coords;
STV_PARTICLE_ID *id;
{
	BULLET B;

	if ( BULLETS == NULL )
		B = BULLETS = create_bullet();
	
	else
	{
		B = BULLETS;

		while ( B->next != NULL )
			B = B->next;

		B = B->next = create_bullet();
	}

	B->loc_x = coords[0];
	B->loc_y = coords[1];

	*id = (STV_PARTICLE_ID) B;
}


void
set_bullet_pfield( id, field, value )
STV_PARTICLE_ID id;
STV_VALUE field;
STV_VALUE value;
{
	BULLET B;

	long fid;

	B = (BULLET) id;

	fid = (long) field;

	switch ( fid )
	{
		case PF_BULLET_BID:
			B->bid = STV_VALUE_OF( value, int );
			break;

		case PF_BULLET_AGE:
			B->age = STV_VALUE_OF( value, int );
			break;

		case PF_BULLET_VEL_ANGLE:
			B->vel_angle = STV_VALUE_OF( value, double );
			break;

		case PF_BULLET_VEL_MAGN:
			B->vel_magn = STV_VALUE_OF( value, double );
			break;

		case PF_BULLET_PARENT:
			B->parent = STV_VALUE_OF( value, int );
			break;

		default:
			break;
	}
}


void
get_explosion( xnum, R, id, coords )
int xnum;
STV_REGION R;
STV_PARTICLE_ID *id;
int *coords;
{
	static EXPLOSION E_save = (EXPLOSION) NULL;
	static int count_save = -1;

	EXPLOSION E;

	int ecoords[2];

	int count;

	/* Reset */

	if ( xnum == -1 )
	{
		*id = (STV_PARTICLE_ID) NULL;

		E_save = (EXPLOSION) NULL;

		count_save = -1;
	}

	/* Restore */

	if ( xnum >= count_save && count_save != -1 )
	{
		E = E_save;

		count = count_save;
	}

	/* Fresh Search */

	else
	{
		E = EXPLOSIONS;

		count = 0;
	}

	/* Search for Explosion */

	while ( E != NULL )
	{
		ecoords[0] = E->loc_x;
		ecoords[1] = E->loc_y;

		if ( stv_particle_in_region( R, ecoords, 2 ) )
		{
			count++;

			if ( count == xnum )
			{
				*id = (STV_PARTICLE_ID) E;

				coords[0] = ecoords[0];
				coords[1] = ecoords[1];

				E_save = E->next;

				count_save = count;

				return;
			}
		}

		E = E->next;
	}

	/* Explosion Not Found */

	*id = (STV_PARTICLE_ID) NULL;

	E_save = (EXPLOSION) NULL;

	count_save = -1;
}


void
get_explosion_pfield( id, field, value )
STV_PARTICLE_ID id;
STV_VALUE field;
STV_VALUE *value;
{
	EXPLOSION E;

	long fid;

	E = (EXPLOSION) id;

	fid = (long) field;

	switch ( fid )
	{
		case PF_EXPLOSION_EID:
			*value = (STV_VALUE) &(E->eid);
			break;

		case PF_EXPLOSION_STATE:
			*value = (STV_VALUE) &(E->state);
			break;

		case PF_EXPLOSION_ANGLE:
			*value = (STV_VALUE) &(E->angle);
			break;

		case PF_EXPLOSION_STRENGTH:
			*value = (STV_VALUE) &(E->strength);
			break;

		case PF_EXPLOSION_RADIUS:
			*value = (STV_VALUE) &(E->radius);
			break;

		default:
			*value = (STV_VALUE) NULL;
			break;
	}
}


void
make_explosion( coords, id )
int *coords;
STV_PARTICLE_ID *id;
{
	EXPLOSION E;

	if ( EXPLOSIONS == NULL )
		E = EXPLOSIONS = create_explosion();
	
	else
	{
		E = EXPLOSIONS;

		while ( E->next != NULL )
			E = E->next;

		E = E->next = create_explosion();
	}

	E->loc_x = coords[0];
	E->loc_y = coords[1];

	*id = (STV_PARTICLE_ID) E;
}


void
set_explosion_pfield( id, field, value )
STV_PARTICLE_ID id;
STV_VALUE field;
STV_VALUE value;
{
	EXPLOSION E;

	long fid;

	E = (EXPLOSION) id;

	fid = (long) field;

	switch ( fid )
	{
		case PF_EXPLOSION_EID:
			E->eid = STV_VALUE_OF( value, int );
			break;

		case PF_EXPLOSION_STATE:
			E->state = STV_VALUE_OF( value, int );
			break;

		case PF_EXPLOSION_ANGLE:
			E->angle = STV_VALUE_OF( value, double );
			break;

		case PF_EXPLOSION_STRENGTH:
			E->strength = STV_VALUE_OF( value, int );
			break;

		case PF_EXPLOSION_RADIUS:
			E->radius = STV_VALUE_OF( value, int );
			break;

		default:
			break;
	}
}


void
normalize_angle( diff, mag )
double *diff;
double *mag;
{
	*mag = fabs( *diff );

	while ( *mag > PI )
	{
		*diff += *diff > 0 ? ( -2.0 * PI ) : ( 2.0 * PI );

		*mag = fabs( *diff );
	}
}


void
rand_point( R, x, y )
REGION R;
int *x, *y;
{
	*x = R->xmin + ( (int) random() % ( R->xmax - R->xmin ) );

	*y = R->ymin + ( (int) random() % ( R->ymax - R->ymin ) );
}


double
rand_double()
{
	double r;

	r = (double) ( (int) random() % 1000000 ) / 1000000.0;

	return( r );
}


double
rand_angle()
{
	double angle;

	angle = ( -1.0 * PI ) + ( 2.0 * PI * rand_double() );
	
	return( angle );
}


double
distance( x1, y1, x2, y2 )
double x1, y1, x2, y2;
{
	return( sqrt(((x2 - x1) * (x2 - x1)) + ((y2 - y1) * (y2 - y1))) );
}


double
angle_of( x1, y1, x2, y2 )
double x1, y1, x2, y2;
{
	double slope;
	double at;

	/* North is heading angle 0.0 */
	/* y's grow downward from the top of the screen */

	if ( x2 == x1 )
	{
		if ( y2 > y1 )
			at = -1.0 * PI;

		else
			at = 0.0;
	}

	else
	{
		slope = (y1 - y2) / (x2 - x1);

		at = atan( slope );

		if ( x1 > x2 )
			at += PI / 2.0;

		else
			at -= PI / 2.0;
	}

	return( at );
}


double
sq( x )
double x;
{
	return( x * x );
}


void
memcheck( ptr, name )
char *ptr;
char *name;
{
	if ( ptr == NULL )
	{
		fprintf( stderr, "\nError Allocating Memory for \"%s\"\n\n",
			name );

		exit( -1 );
	}
}

