
/* $Id: slicer.c,v 1.52 1999/06/25 20:06:56 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


#include "slicer.h"
#include "slicer_glob.h"


/* TCL Initialization Routine */

int Tcl_AppInit();


/* TCL C Command Routines */

int get_slicer_default_dir_proc();
int register_menuvar_proc();
int load_bitmap_file_proc();
int set_plane_choice_proc();
int set_color_range_proc();
int draw_color_map_proc();
int get_field_list_proc();
int set_field_name_proc();
int set_vis_region_proc();
int canvas_handle_proc();
int fix_help_line_proc();
int double_click_proc();
int set_app_name_proc();
int get_app_name_proc();
int set_vis_freq_proc();
int set_vis_mode_proc();
int strip_label_proc();
int resize_stv_proc();
int save_field_proc();
int title_info_proc();
int sleep_set_proc();
int lock_set_proc();
int set_opt_proc();
int pad_proc();


void usage();
void quit_proc();


/* MAIN */

int
main( argc, argv )
int argc;
char **argv;
{
	char **mung_argv;
	int mung_argc;
	int i;

	/* Mung Argv to Circumvent TK Stupidity */

	mung_argc = argc + 1;

	mung_argv = (char **) MALLOC( (STV_SIZE_T) ( mung_argc + 1 )
		* sizeof( char * ) );
	memcheck( mung_argv, "Munged Argv List" );

	mung_argv[0] = argv[0];

	mung_argv[1] = stv_copy_str( "--" );

	for ( i=2 ; i < mung_argc ; i++ )
		mung_argv[i] = argv[ i - 1 ];

	mung_argv[ mung_argc ] = (char *) NULL;

	/* Pass to TK */

	Tk_Main( mung_argc, mung_argv, Tcl_AppInit );

	return( 0 );
}


int Tcl_AppInit( itp )
Tcl_Interp *itp;
{
	/* Save Interpreter */

	interp = itp;

	/* Initialize TCL & TK */

	if ( tcl_init() == TCL_ERROR )
		return( TCL_ERROR );

	/* Get Main Window */

	Top = Tk_MainWindow( interp );

	if ( Top == NULL )
		return( TCL_ERROR );

	Disp = Tk_Display( Top );

	/* Initialize Program Constants & Structs */

	if ( program_init() == TCL_ERROR )
		return( TCL_ERROR );

	/* Handle Command Line Args */

	if ( read_args() == TCL_ERROR )
		return( TCL_ERROR );

	/* Initialize Modules */

	if ( do_stv_init() == TCL_ERROR )
		return( TCL_ERROR );

	if ( window_init() == TCL_ERROR )
		return( TCL_ERROR );

	/* Set Up User-Specific Startup File */

#if ( TCL_MAJOR_VERSION > 7 ) || ( TCL_MINOR_VERSION >= 5 )

	SET_TCL_GLOBAL( interp, "tcl_rcFileName", "~/.slicerrc" );

#else

	tcl_RcFileName = "~/.slicerrc";

#endif

	/* Return to tkMain to Start Tk_MainLoop() */

	return( TCL_OK );
}


int
program_init()
{
	struct passwd *pw;

	char *slicer_dir;
	char *home;

	char tmp[1024];

	int i;

	/* Seed Random Generator */

	srandom( (int) time( (time_t *) NULL ) );

	/* Get Slicer Main Directory */

	slicer_dir = getenv( "STV_ROOT" );

	if ( slicer_dir == NULL )
	{
		sprintf( tmp, "%s", SLICER_DEFAULT_DIR );

		if ( tmp[0] != '/' )
		{
			home = getenv( "HOME" );

			if ( home == NULL )
			{
				if ( (pw = getpwuid( getuid() )) != NULL )
					home = pw->pw_dir;
			
				else
					home = "/";
			}

			sprintf( tmp, "%s/%s/%s", home,
				STV_DEFAULT_DIR, SLICER_DEFAULT_DIR );
		}
	}

	else
		sprintf( tmp, "%s/%s", slicer_dir, SLICER_DEFAULT_DIR );

	SLICER_DIR = stv_copy_str( tmp );

	/* Initialize Tcl / Tk Globals */

	MENUVAR_AUTO_COLOR = (char *) NULL;
	MENUVAR_MIN_MAX = (char *) NULL;
	MENUVAR_CUMULATIVE = (char *) NULL;

	/* Initialize STV Globals */

	SLICER_RADIUS_FIELD = (char *) NULL;
	SLICER_COLOR_FIELD = (char *) NULL;
	SLICER_FIELD_NAME = (char *) NULL;
	SLICER_FIELD_TYPE = "Field";
	SLICER_FIELD = (char *) NULL;
	SLICER_APP = (char *) NULL;

	CURRENT_VFG = (STV_VIEW_FIELD_GROUP) NULL;

	CURRENT_VF = (STV_VIEW_FIELD) NULL;

	CURRENT_VP = (STV_VIEW_PARAM) NULL;

	VIEWER = (STV_VIEWER) NULL;

	stv_init_region( MAX_REGION );

	stv_init_region( VISREGION );

	stv_init_cell( VISCELL );

	STV_CELL_PARTICLE( VISCELL) = 1;

	VIEWER_LOCK = 0;
	VIEWER_KICK = stvFalse;

	INITIALIZED = stvFalse;

	CONNECTED = stvFalse;

	FRAME_PENDING = stvFalse;

	PLANE_CHOICE = PLANE_XY;

	REGIONS = (REGION_STACK) NULL;

	THRASHING = stvFalse;
	THRASH_FACTOR = 1;

	AUTO_COLOR_MAP = stvFalse;

	SHOW_MIN_MAX = stvFalse;

	SHOW_CUMULATIVE = stvFalse;

	CUMULATIVE_MIN = 0.0;
	CUMULATIVE_MAX = 0.0;
	CUMULATIVE_SET = 0;

	STEERING = stvFalse;

	VIS_MODE = VIS_MODE_NORMAL;

	PSCALE = -1.0;

	SCALE = -1;

	NX = NY = -1;

	XON_KICK = stvFalse;

	SLEEP = stvFalse;

	NSTEPS = 0;

	APP_LOCKED = APP_LOCK_FREE;

	REALLOC = stvTrue;

	RESTART = stvFalse;

	FREQ = 1;

	/* Initialize Slicer Globals */

	MAIN_WIN = (WIN) NULL;

	DIFF_FIELD[0] = (float *) NULL;
	DIFF_FIELD[1] = (float *) NULL;

	DIFF_FIELD_NELEMS = -1;

	DIFF_INDEX = -1;

	SAVE_FIELD = (float *) NULL;

	SAVE_FIELD_NELEMS = -1;

	FIELD_NELEMS = -1;

	COLOR_MIN_VALUE = 0.0;
	COLOR_MAX_VALUE = 100.0;

	COLORIDS = (int *) NULL;
	NCOLORIDS = 0;

	RECTIDS = (RECT) NULL;
	RECTIDSIZE = 0;
	NRECTIDS = 0;

	CIRCLEIDS = (CIRCLE) NULL;
	CIRCLEIDSIZE = 0;
	NCIRCLEIDS = 0;

	BORDERID = (RECT) MALLOC( sizeof(struct rect_struct) );
	memcheck( BORDERID, "Border Rectangle ID" );

	BORDERID->id = -1;

	STEER_RELEASE = stvTrue;

	STEER_ACTIVE = stvFalse;

	PTR_VALID = stvFalse;

	PTR_X = -1;
	PTR_Y = -1;

	REGION_ID = -1;

	vflag = 0;

	stv_setopt( stvOptNull, stvOptVerbose, vflag ? STV_DEBUG_ALL : 0 );

	return( TCL_OK );
}


read_args()
{
	char tmp[1024];

	char *argv_str;

	int i, j, k;
	int do_usage;
	int len;

	argv_str = GET_TCL_GLOBAL( interp, "argv" );

	if ( Tcl_SplitList( interp, argv_str, &Argc, &Argv ) == TCL_ERROR )
		return( TCL_ERROR );

	do_usage = 0;

	for ( i=0 ; i < Argc ; i++ )
	{
		if ( Argv[i][0] == '-' )
		{
			k = i + 1;

			len = strlen( Argv[i] );

			for ( j=1 ; j < len ; j++ )
			{
				switch ( Argv[i][j] )
				{
					case 'A':
					{
						if ( SLICER_APP != NULL )
							FREE( SLICER_APP );

						if ( k < Argc )
							SLICER_APP = stv_copy_str( Argv[ k++ ] );

						else
							do_usage++;

						break;
					}

					case 'F':
					{
						if ( k < Argc )
							FREQ = atoi( Argv[ k++ ] );

						else
							do_usage++;

						break;
					}

					case 'V':
					{
						if ( k < Argc )
						{
							if ( !strcmp( Argv[ k ], "normal" ) )
								VIS_MODE = VIS_MODE_NORMAL;

							else if ( !strcmp( Argv[ k ], "diff" ) )
								VIS_MODE = VIS_MODE_DIFF;

							else if ( !strcmp( Argv[ k ], "diffsave" ) )
								VIS_MODE = VIS_MODE_DIFFSAVE;

							else
							{
								printf( "Unknown Vis Mode \"%s\".\n",
									Argv[ k ] );
							
								do_usage++;
							}
						}

						else
							do_usage++;

						k++;

						break;
					}

					case 'C':
					{
						if ( k < Argc )
							AUTO_COLOR_MAP = atoi( Argv[ k++ ] );

						else
							do_usage++;

						break;
					}

					case 'M':
					{
						if ( k < Argc )
							SHOW_MIN_MAX = atoi( Argv[ k++ ] );

						else
							do_usage++;

						break;
					}

					case 'T':
					{
						if ( k < Argc )
							SHOW_CUMULATIVE = atoi( Argv[ k++ ] );
						
						else
							do_usage++;

						break;
					}

					case 'v':
					{
						vflag++;

						stv_setopt( stvOptNull, stvOptVerbose,
							vflag ? STV_DEBUG_ALL : 0 );

						break;
					}

					default:
					{
						printf( "Unknown Option -%c\n", Argv[i][j] );

						do_usage++;

						break;
					}
				}
			}

			i = k - 1;
		}
	}

	if ( do_usage )
		usage();

	return( TCL_OK );
}


void
usage()
{
	printf( "\nusage:  slicer [-A <app name>] [-V <mode>] " );
	printf( "[-[CMT] <0|1>] [-F <freq>] [-v]\n\n" );

	printf( "where:\n" );
	printf( "------\n" );

	printf( "-A <appname> = Set Application Name.\n" );

	printf( "-V <mode>    = Set Visualization Mode:\n" );
	printf( "                <mode> =\n" );
	printf( "                   \"normal\"   - Normal Mode.\n" );
	printf( "                   \"diff\"     - Differential Mode.\n" );
	printf( "                   \"diffsave\" - Diff-to-Saved Mode.\n" );

	printf( "-C <0|1>     = Set Auto Color Map Range (0=off,1=on).\n" );

	printf( "-M <0|1>     = Show Min & Max Data Values (0=off,1=on).\n" );

	printf( "-T <0|1>     = SHow Cumulative Min & Max (0=off,1=on).\n" );

	printf( "-F <freq>    = Set STV Frame Frequency.\n" );

	printf( "-v           = Verbose Operation.\n" );

	printf( "\n" );

	exit( 0 );
}


do_stv_init()
{
	int cc;

	cc = stv_myid( &MYTID );

	if ( !cc )
	{
		stv_mp_err( "stv_myid" );

		return( TCL_ERROR );
	}

	return( TCL_OK );
}


void
do_stv_cleanup()
{
	CURRENT_VFG = (STV_VIEW_FIELD_GROUP) NULL;

	/* Unset Flags */

	INITIALIZED = stvFalse;

	CONNECTED = stvFalse;

	STEERING = stvFalse;

	/* Reset to Slice Mode */

	if ( !RESTART )
	{
		stv_init_region( MAX_REGION );

		stv_init_region( VISREGION );

		stv_init_cell( VISCELL );

		SET_TCL_GLOBAL( interp, "plane_choice", "XY" );

		safe_eval( interp, "layout_main_panel" );

		safe_eval( interp, "set_plane_choice XY" );
	}

	/* Sleep to Fix Nasty Race Condition. */
	/* NEVER USE THIS */

	stvSleep( 1 );
}


tcl_init()
{
	/* Initialize TCL / TK */

	if ( Tcl_Init( interp ) == TCL_ERROR )
		return( TCL_ERROR );

	if ( Tk_Init( interp ) == TCL_ERROR )
		return( TCL_ERROR );

	/* Create TCL Commands for Action Routines */

	Tcl_CreateCommand( interp, "canvas_handle",
		canvas_handle_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "set_plane_choice",
		set_plane_choice_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "set_color_range",
		set_color_range_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "draw_color_map",
		draw_color_map_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "get_field_list",
		get_field_list_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "set_field_name",
		set_field_name_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "set_vis_region",
		set_vis_region_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "set_vis_freq",
		set_vis_freq_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "set_vis_mode",
		set_vis_mode_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "set_app_name",
		set_app_name_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "get_app_name",
		get_app_name_proc, (ClientData) NULL, (vfp) NULL );

	/* Create TCL Utility Commands */

	Tcl_CreateCommand( interp, "get_slicer_default_dir",
		get_slicer_default_dir_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "register_menuvar",
		register_menuvar_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "load_bitmap_file",
		load_bitmap_file_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "fix_help_line",
		fix_help_line_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "double_click",
		double_click_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "strip_label",
		strip_label_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "resize_stv",
		resize_stv_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "save_field",
		save_field_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "title_info",
		title_info_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "sleep_set",
		sleep_set_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "lock_set",
		lock_set_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "set_opt",
		set_opt_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "viewer_lock",
		(Tcl_CmdProc *) viewer_lock, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "viewer_unlock",
		(Tcl_CmdProc *) viewer_unlock, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "quit",
		(Tcl_CmdProc *) quit_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "pad",
		pad_proc, (ClientData) NULL, (vfp) NULL );

	return( TCL_OK );
}


window_init()
{
	static char fname[1024];

	FILE *fptest;

	char *tmp;

	/* Create Interface - Run TCL Script */

	fptest = fopen( "slicer.tcl", "r" );

	if ( fptest != NULL )
	{
		strcpy( fname, "slicer.tcl" );

		fclose( fptest );
	}

	else
		sprintf( fname, "%s/slicer.tcl", SLICER_DIR );

	if ( Tcl_EvalFile( interp, fname ) == TCL_ERROR )
		return( TCL_ERROR );

	/* Read in TCL Global Constants */

	SL = GET_TCL_GLOBAL( interp, "SL" );
	SL_C = GET_TCL_GLOBAL( interp, "SL_C" );
	SL_M = GET_TCL_GLOBAL( interp, "SL_M" );

	/* Get TCL Spacing Constants */

	tmp = GET_TCL_GLOBAL( interp, "border_space" );
	BORDER_SPACE = atoi( tmp );

	tmp = GET_TCL_GLOBAL( interp, "frame_border" );
	FRAME_BORDER = atoi( tmp );

	tmp = GET_TCL_GLOBAL( interp, "row_height" );
	ROW_HEIGHT = atoi( tmp );

	tmp = GET_TCL_GLOBAL( interp, "col_width" );
	COL_WIDTH = atoi( tmp );

	/* Initialize Window Display Objects */

	MAIN_WIN = create_win();

	MAIN_WIN->canvas = SL_C;

	MAIN_WIN->htvar = "slicer_height";
	MAIN_WIN->wtvar = "slicer_width";

	MAIN_WIN->xoff = MAIN_WIN->yoff = 0;

	/* Set Option Default Menuvars */

	if ( MENUVAR_AUTO_COLOR )
		SET_TCL_GLOBAL( interp, MENUVAR_AUTO_COLOR,
			( AUTO_COLOR_MAP ) ? "ON" : "OFF" );

	if ( MENUVAR_MIN_MAX )
		SET_TCL_GLOBAL( interp, MENUVAR_MIN_MAX,
			( SHOW_MIN_MAX ) ? "ON" : "OFF" );

	if ( MENUVAR_CUMULATIVE )
		SET_TCL_GLOBAL( interp, MENUVAR_CUMULATIVE,
			( SHOW_CUMULATIVE ) ? "ON" : "OFF" );

	if ( Tcl_Eval( interp, "layout_main_panel" ) == TCL_ERROR )
		return( TCL_ERROR );

	/* Set Up Viewer Handler */

	Tk_DoWhenIdle( viewer_handle, (ClientData) NULL );

	return( TCL_OK );
}


/* ARGSUSED */
int
get_slicer_default_dir_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	char tmp[1024];

	sprintf( tmp, "%s/%s", STV_DEFAULT_DIR, SLICER_DEFAULT_DIR );

	Tcl_SetResult( itp, stv_copy_str( tmp ), TCL_DYNAMIC );

	return( TCL_OK );
}

