
/* $Id: checkers.c,v 3.13 1999/11/03 00:32:48 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/*
	checkers.c - A small test program to test multi-explicit
	decompositions with CUMULVS.
*/


#include <stdio.h>
#include <math.h>
#include "pvm3.h"
#include "stv.h"


/* ---  Adjustable Parameters  --- */


/* The number of processes used is ( COLS * ROWS * SLABS ) */

#define COLS		2
#define ROWS		2
#define SLABS		2

/*
	Note: the size of the whole computational region is 
	(COLS * I_DIM) X (ROWS * J_DIM) X (SLABS * K_DIM)
*/


/* Size of Global Dimensions */

#define MAXI		20
#define MAXJ		40
#define MAXK		80


/* --- Row/Col/Slab of Index Macros */

#define COL_OF( _i )	( (_i) % COLS )
#define ROW_OF( _i )	( ( (_i) / COLS ) % ROWS )
#define SLAB_OF( _i )	( (_i) / ( COLS * ROWS ) )


/* --- Miscellaneous Parameters --- */

#define CHECKERSGRP		"checkers"
#define STVTAG			100
#define NEIGHMSG		1001
#define NTASKS			( COLS * ROWS * SLABS )
#define NFIELDS			1
#define NPARAMS			4

#define NeighMsgMarker	-1


/* --- Other Options --- */

/* #define INTERMEDIATE_SEND */

/* #define VERBOSE    stvTrue */
#define VERBOSE    stvFalse


void dump_params_changed();


/* --- Main Program --- */

void
main()
{

/* --  Variable declarations  -- */

	float *cube;
	float *iptr, *jptr;

	float rate; 
	float value;

	int axisInfo[3];
	int axisInfo2[3];
	int axisType[3];
	int bufid;
	int cc;
	int declare[3];
	int decompId[NFIELDS];
	int fieldId[NFIELDS];
	/* int globali, globalj, globalk; */
	int glb[3], gub[3];
	int info;
	int i, j, k;
	int idim, jdim, kdim;
	int msrc;
	int mycol, myrow, myslab;
	int myinum, mytid;
	int nactual;
	int neigh[6];
	int nspawn;
	int onum, otid;
	int paramIds[NPARAMS];
	int prank, pshape[3], paddr[3];
	int sequence;
	int sizei, sizej, sizek;
	int tids[NTASKS];
	int tmp;

/* --  PVM Initialization  -- */

	mytid = pvm_mytid();
	myinum = pvm_joingroup( CHECKERSGRP );

	printf( "Checkers Started:  Task=%x  Instance=%d\n",
		mytid, myinum );
	fflush( stdout );

/*  --- Should I spawn other tasks? */

	if ( myinum == 0 )
	{ 
		if ( NTASKS - 1 > 0 ) 
		{
			pvm_catchout( stdout );

			nspawn = pvm_spawn( "checkers", (char **) 0,
					PvmTaskDefault,
					/* PvmTaskDebug, */
					"", NTASKS - 1, tids );

			if ( nspawn < NTASKS - 1 )
			{
				printf( "Spawn failed with nspawn = %d\n", nspawn ); 

				printf( "Killing orphaned tasks...\n" );

				for ( i=0 ; i < nspawn ; i++ ) 
					pvm_kill( tids[i] );

				printf( "Failed spawns returned codes:\n" );

				for ( i=nspawn ; i < NTASKS - 1 ; i++ ) 
					printf( "Info was: %d\n", tids[i] );	

				fflush( stdout );

				pvm_exit();

				exit( -1 );
			}
		}
	}

/*  --- Make the group static -- this is also a barrier  */

	pvm_freezegroup( CHECKERSGRP, NTASKS );

/*  --- Determine neighbors for messages */

	mycol = COL_OF( myinum );
	myrow = ROW_OF( myinum );
	myslab = SLAB_OF( myinum );

	/* Left */
	neigh[0] = myinum - 1;
	if ( ROW_OF( neigh[0] ) != myrow || SLAB_OF( neigh[0] ) != myslab
			|| neigh[0] < 0 )
		neigh[0] = -1;

	/* Right */
	neigh[1] = myinum + 1;
	if ( ROW_OF( neigh[1] ) != myrow || SLAB_OF( neigh[1] ) != myslab
			|| neigh[1] >= NTASKS )
		neigh[1] = -1;

	/* Up */
	neigh[2] = myinum - COLS;
	if ( COL_OF( neigh[2] ) != mycol || SLAB_OF( neigh[2] ) != myslab
			|| neigh[2] < 0 )
		neigh[2] = -1;

	/* Down */
	neigh[3] = myinum + COLS;
	if ( COL_OF( neigh[3] ) != mycol || SLAB_OF( neigh[3] ) != myslab
			|| neigh[3] >= NTASKS )
		neigh[3] = -1;

	/* Front */
	neigh[4] = myinum - ( ROWS * COLS );
	if ( COL_OF( neigh[4] ) != mycol || ROW_OF( neigh[4] ) != myrow
			|| neigh[4] < 0 )
		neigh[4] = -1;

	/* Back */
	neigh[5] = myinum + ( ROWS * COLS );
	if ( COL_OF( neigh[5] ) != mycol || ROW_OF( neigh[5] ) != myrow
			|| neigh[5] >= NTASKS )
		neigh[5] = -1;

	nactual = 0;

	for ( i=0 ; i < 6 ; i++ )
	{
		if ( neigh[i] >= 0 )
		{
			neigh[i] = pvm_gettid( CHECKERSGRP, neigh[i] );

			if ( neigh[i] < 0 )
			{
				printf( "Whoa!  Can't Find Neighbor #%d!\n", i );
				fflush( stdout );

				exit( -1 );
			}

			/* notify me if my src task fails */
			cc = pvm_notify( PvmTaskExit, NEIGHMSG, 1, &(neigh[i]) );

			if ( cc < 0 )
			{
				printf( "Error in pvm_notify() - Bailing...\n" );
				fflush( stdout );

				exit( -1 );
			}

			nactual++;
		}

		if ( VERBOSE )
		{
			printf( "neigh[%d] = %x\n", i, neigh[i] );
			fflush( stdout );
		}
	}

/* --- Initialize CUMULVS steering and visualization */

	stv_setopt( stvOptNull, stvOptVerbose,
		VERBOSE ? STV_DEBUG_ALL : 0 );

	/* stv_setopt( stvOptNull, stvOptAppLock, stvTrue ); */

	stv_init( CHECKERSGRP, STVTAG, NTASKS, myinum );

/* --- Record the size of the whole computational region */

	glb[0] = 1; glb[1] = 1; glb[2] = 1;

	gub[0] = MAXI;
	gub[1] = MAXJ;
	gub[2] = MAXK;
		
/* --- Initialize the parameters */

	rate = 0.1;

	sizei = 4;
	sizej = 6;
	sizek = 8;

	if ( myinum == 0 ) 
	{
		printf( "sizes: i=%d j=%d k=%d \n", sizei, sizej, sizek );
		fflush( stdout );
	}

/* --- Define the steering parameters */

	paramIds[0] = stv_paramDefine( "Rate", (void *) &rate,
			stvFloat, stvVisOnly );

	paramIds[1] = stv_paramDefine( "Size I", (void *) &sizei,
			stvInt, stvVisOnly ); 

	paramIds[2] = stv_paramDefine( "Size J", (void *) &sizej,
			stvInt, stvVisOnly ); 

	paramIds[3] = stv_paramDefine( "Size K", (void *) &sizek,
			stvInt, stvVisOnly ); 

/* --- Declare a decomposition of density */

	/* -- it is multi-explicit in all axes */
	axisType[0] = stvExplicit;
	axisType[1] = stvExplicit;
	axisType[2] = stvExplicit;

	/* First Explicit Region */
	i = ( mycol * sizei ) + glb[0];
	j = ( myrow * sizej ) + glb[1];
	k = ( myslab * sizek ) + glb[2];

	/* Region Bounds */
	axisInfo[0] = i;
	axisInfo2[0] = i + sizei - 1;
	axisInfo[1] = j;
	axisInfo2[1] = j + sizej - 1;
	axisInfo[2] = k;
	axisInfo2[2] = k + sizek - 1;

	/* Process Topology */
	prank = 3;
	pshape[0] = COLS;
	pshape[1] = ROWS;
	pshape[2] = SLABS;

	decompId[0] = stv_decompDefine( 3, axisType, axisInfo, axisInfo2,
			glb, gub, prank, pshape );

	if ( VERBOSE )
	{
		printf( "decompId[0] = %x\n", decompId[0] );
		fflush( stdout );
	}

/* --- Define Remaining Explicit Regions */

	idim = ( sizei <= MAXI ) ? sizei : MAXI;

	do
	{
		i += sizei * COLS;

		if ( i <= MAXI )
		{
			axisInfo[0] = i;
			axisInfo2[0] = i + sizei - 1;

			if ( axisInfo2[0] > MAXI )
				axisInfo2[0] = MAXI;

			if ( (cc = stv_decompAddExplicitBounds( decompId[0],
					0, axisInfo[0], axisInfo2[0] )) != stvStatusOk )
			{
				printf( "Oops... Couldn't Add Explicit Bounds (I)!\n" );
				fflush( stdout );

				exit( -1 );
			}

			idim += ( i + sizei <= MAXI ) ? sizei : ( MAXI - i + 1 );
		}
	}
	while ( i <= MAXI );

	jdim = ( sizej <= MAXJ ) ? sizej : MAXJ;

	do
	{
		j += sizej * ROWS;

		if ( j <= MAXJ )
		{
			axisInfo[1] = j;
			axisInfo2[1] = j + sizej - 1;

			if ( axisInfo2[1] > MAXJ )
				axisInfo2[1] = MAXJ;

			if ( (cc = stv_decompAddExplicitBounds( decompId[0],
					1, axisInfo[1], axisInfo2[1] )) != stvStatusOk )
			{
				printf( "Oops... Couldn't Add Explicit Bounds (J)!\n" );
				fflush( stdout );

				exit( -1 );
			}

			jdim += ( j + sizej <= MAXJ ) ? sizej : ( MAXJ - j + 1 );
		}
	}
	while ( j <= MAXJ );

	kdim = ( sizek <= MAXK ) ? sizek : MAXK;

	do
	{
		k += sizek * SLABS;

		if ( k <= MAXK )
		{
			axisInfo[2] = k;
			axisInfo2[2] = k + sizek - 1;

			if ( axisInfo2[2] > MAXK )
				axisInfo2[2] = MAXK;

			if ( (cc = stv_decompAddExplicitBounds( decompId[0],
					2, axisInfo[2], axisInfo2[2] )) != stvStatusOk )
			{
				printf( "Oops... Couldn't Add Explicit Bounds (K)!\n" );
				fflush( stdout );

				exit( -1 );
			}

			kdim += ( k + sizek <= MAXK ) ? sizek : ( MAXK - k + 1 );
		}
	}
	while ( k <= MAXK );

/* --- allocate local storage for cube */

	printf( "idim=%d jdim=%d kdim=%d\n", idim, jdim, kdim );
	fflush( stdout );

	cube = (float *) malloc( (unsigned) ( idim * jdim * kdim )
		* sizeof( float ) );
	
	if ( cube == NULL )
	{
		printf( "Error Allocating Cube Array...  Bailing Out...\n" );
		fflush( stdout );

		exit( -1 );
	}

/* --- declare the cube data field with the above decomposition */

	paddr[0] = mycol;
	paddr[1] = myrow;
	paddr[2] = myslab;

	declare[0] = idim;
	declare[1] = jdim;
	declare[2] = kdim;

	fieldId[0] = stv_fieldDefine( cube, "cube", decompId[0],
			stvNoFieldOffsets, declare, stvFloat, paddr, stvVisOnly );

	if ( VERBOSE )
	{
		printf( "fieldId[0] = %x\n", fieldId[0] );
		fflush( stdout );
	}

	/* set simulation time step for density field to 1 nanosec */

	stv_setopt( fieldId[0], stvOptFieldTimeScale,
			stv_time_of( 1, stvTimeNanoSec ) );

	/* set base time to -1 secs - count up to 0 (just for fun) */

	stv_setopt( fieldId[0], stvOptSetFieldTime,
			stv_time_of( -1, stvTimeSec ) );

/* --- Verify Local Field Decomposition */

	stv_dumpLocalFieldBounds( fieldId[0], stvVerbose );

	stv_dumpLocalFieldBounds( fieldId[0], stvTerse );

/* ---  Initialize Computational Region */

	for( i=0 ; i < idim ; i++ )
	{
		iptr = cube + ( i * ( jdim * kdim ) );

		for ( j=0 ; j < jdim ; j++ )
		{
			jptr = iptr + ( j * kdim );

			for ( k=0 ; k < kdim ; k++ )	
				jptr[k] = -2;
		}
	}

/* ---  Top of main computational loop */

	sequence = 0;

	for( ; ; )
	{
		if ( myinum == 0 ) 
		{
			printf( "Sequence %d\n", sequence );
			fflush( stdout );
		}

		for( i=0 ; i < idim ; i++ )
		{
			iptr = cube + ( i * ( jdim * kdim ) );

#ifdef INTERMEDIATE_SEND
			/*  --- send intermediate data to any attached viewers */
			stv_sendReadyData( stvSendNoFieldTimeUpdates );
#endif

			for ( j=0 ; j < jdim ; j++ )
			{
				jptr = iptr + ( j * kdim );

				for ( k=0 ; k < kdim ; k++ )	
				{
					value = (float) myinum;

					jptr[k] = value;
				}
			}
		}

			/*
			globali = ( mycol * sizei ) + ( i * sizei * COLS );
				globalj = ( myrow * sizej ) + ( j * sizej * ROWS );
					globalk = ( myslab * sizek )
							+ ( k * sizek * SLABS );
			*/

/*  --- send a message to neighbors
			Keep the tasks pseudo-synched */

		pvm_initsend( PvmDataDefault );
		tmp = NeighMsgMarker;
		pvm_pkint( &tmp, 1, 1 );
		pvm_pkint( &myinum, 1, 1 );
		pvm_pkint( &mytid, 1, 1 );

		for ( i=0 ; i < 6 ; i++ )
		{
			if ( neigh[i] > 0 )
			{
				if ( VERBOSE )
				{
					printf( "Sending message to %x tag=%d\n",
						neigh[i], NEIGHMSG );
					fflush( stdout );
				}

				pvm_send( neigh[i], NEIGHMSG );
			}
		}

		for ( i=0 ; i < nactual ; i++ )
		{
			if ( VERBOSE )
			{
				printf( "Blocking on recv tag=%d\n", NEIGHMSG );
				fflush( stdout );
			}

			bufid = pvm_recv( -1, NEIGHMSG );

			pvm_bufinfo( bufid, (int *) NULL, (int *) NULL, &msrc );

			if ( VERBOSE )
			{
				printf( "Message recvd from %x.\n", msrc );
				fflush( stdout );
			}

			pvm_upkint( &tmp, 1, 1 );

			if ( tmp == NeighMsgMarker )
			{
				for ( j=0 ; j < 6 ; j++ )
				{
 					if ( msrc == neigh[j] )
					{
						pvm_upkint( &onum, 1, 1 );
						pvm_upkint( &otid, 1, 1 );

						if ( VERBOSE )
						{
							printf( "Seq %d: inst=%d tid=%x.\n",
									sequence, onum, otid );
							fflush( stdout );
						}
					}
				}
			}

			else
			{
				printf( "Notify Message...  Task %x Died.\n", tmp );
				fflush( stdout );

				for ( j=0 ; j < 6 ; j++ )
				{
 					if ( tmp == neigh[j] )
					{
						printf( "Task Was A Neighbor - Bail!\n" );
						fflush( stdout );

						exit( 0 );
					}
				}
			}
		}

/*  --- send messages to any attached CUMULVS front-end viewers */

		if ( VERBOSE )
		{
			printf( "Calling stv_sendReadyData()...\n" );
			fflush( stdout );
		}	

		info = stv_sendReadyData( stvSendDefault );

		if ( VERBOSE )
		{
			printf( "info = %d\n", info );
			fflush( stdout );
		}

		if ( info > 0 && myinum == 0 ) 
		{
			dump_params_changed( info, paramIds,
					rate, sizei, sizej, sizek );
		}

		stvSleep( 1 );

		sequence++;
	}
}


/* print out message that parameters have been updated by a viewer */

void
dump_params_changed( info, paramIds, rate, sizei, sizej, sizek )
int info;
int *paramIds;
float rate;
int sizei;
int sizej;
int sizek;
{
	int i;

	printf( "stv_sendReadyData returned %d\n", info );

	for ( i=0 ; i < NPARAMS ; i++ )
	{
		if( stv_isParamChanged( paramIds[i] ) )
		{
		 	if ( i == 0 )
				printf( "Parameter 'Rate' changed to %f\n", rate );
			else if ( i == 1 )	 	
				printf( "Parameter 'Size I' changed to %d\n", sizei );
			else if ( i == 2 )	 	
				printf( "Parameter 'Size J' changed to %d\n", sizej );
			else if ( i == 3 )	 	
				printf( "Parameter 'Size K' changed to %d\n", sizek );
		}
	}
}

