
/* $Id: cplane.c,v 3.9 1999/06/18 18:16:49 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/*
	cplane.c - A better test program to test the packing
	setup with CUMULVS.
*/


#include <stdio.h>
#include <math.h>

#include "pvm3.h"

#include "stv.h"


/* ---  Adjustable Parameters  --- */


/* Size of Global Computational Region is I_DIM X J_DIM X K_DIM */

#define NDIM	3

#define IDIM	20
#define JDIM	30
#define KDIM	40


/* Size of Array Allocation */

#define IALLOC	100
#define JALLOC	100
#define KALLOC	100


/* Misc Constants */

#define CPLANEGRP	"cplane"

#define STVTAG		100

#define RINGMSG		1001

#define CYCLESIZE	3


/* The processor shape used */

#define MAXPROCDIM	3

int PSHAPE[]	=	{ 2, 3, 2 };

int PRANK		=	sizeof( PSHAPE ) / sizeof( PSHAPE[0] );

int NTASKS;


/* Explicit Axis Munging Function Macro */

#define MUNG( _n )	( (_n) * ( (_n) + 1 ) / 2 )


/* Routines */

void	bail();
void	cumulvs_init();
void	program_init();
void	pvm_init();


/* Computational Parameters */

float CENTERI;
float CENTERJ;
float CENTERK;

float INTERCEPT;

float SLOPE;

float RISE;

float FACTOR;
float BASE;

int OFFSETI;
int OFFSETJ;
int OFFSETK;

int RESET;


/* Other Flags */

int vflag;


/* --- Main Program --- */

main()
{
	float density[IALLOC][JALLOC][KALLOC];

	float x, y, z;
	float value;
	float X, Z;
	float sgn;

	int glb[NDIM], gub[NDIM];
	int paddr[MAXPROCDIM];

	int paramId[10];
	int decompId;
	int fieldId;

	int globali, globalj, globalk;
	int locali, localj, localk;
	int dsttid, srctid;
	int begink, endk;
	int sequence;
	int i, j, k;
	int bstart;
	int cstart;
	int myinst;
	int block;
	int bufid;
	int cstep;
	int mytid;
	int slice;
	int info;
	int msrc;
	int qcnt;
	int q;

	/* Program Initialization */

	program_init();

	/* PVM Initialization */

	pvm_init( &mytid, &myinst, &dsttid, &srctid );

	/* Record the size of the computational region */

	glb[0] = -IDIM / 2;
	glb[1] = -JDIM / 2;
	glb[2] = -KDIM / 2;

	gub[0] = IDIM / 2;
	gub[1] = JDIM / 2;
	gub[2] = KDIM / 2;

	/* Determine My Processor Address */

	paddr[0] = myinst / ( PSHAPE[1] * PSHAPE[2] );
	paddr[1] = ( myinst / PSHAPE[2] ) % PSHAPE[1];
	paddr[2] = myinst % PSHAPE[2];

	/* Determine Local Bounds for Explicit Axis */

	/* Axis #1 */

	block = (int) ceil( (double) ( gub[0] - glb[0] + 1 )
		/ (double) PSHAPE[0] );

	bstart = block * paddr[0];

	if ( vflag )
		printf( "block=%d  bstart=%d\n", block, bstart );

	/* Axis #2 */

	cstart = CYCLESIZE * paddr[1];

	cstep = CYCLESIZE * PSHAPE[1];

	if ( vflag )
		printf( "cstart=%d  cstep=%d\n", cstart, cstep );

	/* Axis #3 */

	slice = ( ( gub[2] - glb[2] + 1 ) - MUNG( PSHAPE[2] ) ) / PSHAPE[2];

	begink = glb[2] + ( paddr[2] * slice ) + MUNG( paddr[2] );

	endk = begink + slice + paddr[2];

	endk = ( endk > gub[2] ) ? gub[2] : endk;

	if ( vflag )
		printf( "slice=%d  begink=%d  endk=%d\n", slice, begink, endk );

	/* CUMULVS Initialization */

	cumulvs_init( myinst, density, glb, gub, paddr, begink, endk,
		&decompId, &fieldId, paramId );

	/* Top of main computational loop */

	sequence = 1;

	for ( ; ; )
	{
		if ( !myinst )
			printf( "Sequence %d\n", sequence );

		for ( i=0 ; i < block ; i++ )
		{
			globali = glb[0] + bstart + i;

			locali = OFFSETI + i;

			q = cstart;

			qcnt = 0;

			do
			{
				for ( j=0 ; j < CYCLESIZE ; j++ )
				{
					globalj = glb[1] + q + j;

					localj = OFFSETJ + ( qcnt * CYCLESIZE ) + j;

					if ( globalj <= gub[1] )
					{
						for ( k=0 ; k < ( endk - begink + 1 ) ; k++ )
						{
							globalk = begink + k;

							localk = OFFSETK + k;

							x = (float) globali - CENTERI;
							y = (float) globalj - CENTERJ;
							z = (float) globalk - CENTERK;

							X = ( x + y ) / 2.0;

							Z = z - ( ( SLOPE * X ) + INTERCEPT );

							sgn = ( Z >= 0 ) ? 1.0 : -1.0;

							value = RISE * sgn * (float) sequence
								* (float) pow( (double) BASE,
									(double) FACTOR
										* fabs( (double) Z ) );

							if ( vflag )
							{
								printf( "density[%d][%d][%d] = ",
									locali, localj, localk );

								printf( "f(%f,%f,%f) = %f\n",
									x, y, z, value );
							}

							density[ locali ][ localj ][ localk ] =
								value;
						}
					}
				}

				q += cstep;

				qcnt++;
			}
			while ( q < gub[1] - glb[1] + 1 );
		}

		/*  --- send a message to mod( myinum + 1, ntasks ).  */
		/*  Keep the tasks pseudo-synched.  */

		pvm_initsend( PvmDataDefault );

		pvm_pkint( &myinst, 1, 1 );
		pvm_pkint( &mytid, 1, 1 );

		pvm_send( dsttid, RINGMSG );

		bufid = pvm_recv( -1, RINGMSG );

		pvm_bufinfo( bufid, (int *) NULL, (int *) NULL, &msrc );

 		if ( msrc != srctid )
		{
			printf(
				"Ring message seq %d from %x instead of %x, exiting.\n",
					sequence, msrc, srctid );

			pvm_bcast( CPLANEGRP, RINGMSG );

			bail();
		}

		/* send messages to any attached CUMULVS front-end viewers */

		info = stv_sendToFE();

		if ( info > 0 && !myinst )
			printf( "stv sendToFE returned %d\n", info );

		stvSleep( 1 );

		if ( RESET )
		{
			sequence = 1;

			RESET = 0;
		}
		
		else
			sequence++;
	}
}


void
program_init()
{
	int i;

	NTASKS = 1;

	for ( i=0 ; i < PRANK ; i++ )
		NTASKS *= PSHAPE[i];

	CENTERI = 0.0;
	CENTERJ = 0.0;
	CENTERK = 0.0;

	INTERCEPT = -3.0;

	SLOPE = 1.2;

	RISE = 2.0;

	FACTOR = -0.75;
	BASE = 1.1;

	OFFSETI = 10;
	OFFSETJ = 20;
	OFFSETK = 30;

	RESET = 0;

	vflag = stvFalse;
}


void
pvm_init( mytid, myinst, dsttid, srctid )
int *mytid;
int *myinst;
int *dsttid;
int *srctid;
{
	int tids[100];

	int dst, src;
	int nspawn;
	int i;

	*mytid = pvm_mytid();

	*myinst = pvm_joingroup( CPLANEGRP );

	/* Should I spawn other tasks? */

	if ( !(*myinst) )
	{
		if ( NTASKS > 1 )
		{
			pvm_catchout( stdout );

			nspawn = pvm_spawn( "cplane", (char **) 0,
					PvmTaskDefault,
					/* PvmTaskDefault | PvmTaskDebug, */
					"", NTASKS - 1, tids );

			if ( nspawn < NTASKS - 1 )
			{
				printf( "Spawn failed with nspawn = %d\n", nspawn );

				printf( "Killing orphaned tasks...\n" );

				for ( i=0 ; i < nspawn ; i++ )
				{
					if ( tids[i] < 0 )
					{
						printf( "Task #%d: error = %d\n",
							i + 1, tids[i] );
					}

					else
						pvm_kill( tids[i] );
				}

				bail();
			}
		}
	}

	/* Make the group static -- this is also a barrier */

	pvm_freezegroup( CPLANEGRP, NTASKS );

	/* Determine src and destination for ring message */

	if ( (dst = (*myinst) + 1) == NTASKS )
		dst = 0;

	*dsttid = pvm_gettid( CPLANEGRP, dst );

	if ( (src = (*myinst) - 1) < 0 )
		src = NTASKS - 1;

	*srctid = pvm_gettid( CPLANEGRP, src );

	/* notify me if my src task fails */

	pvm_notify( PvmTaskExit, RINGMSG, 1, srctid );
}


void
cumulvs_init( myinst, density, glb, gub, paddr, begink, endk,
	decompId, fieldId, paramId )
int myinst;
float *density;
int *glb;
int *gub;
int *paddr;
int begink;
int endk;
int *decompId;
int *fieldId;
int *paramId;
{
	void *paramArray[2];
	void *paramPtr;

	char *paramNames[2];
	char *paramName;

	int axisInfo1[NDIM];
	int axisInfo2[NDIM];
	int axisType[NDIM];
	int declare[NDIM];
	int offsets[NDIM];

	int paramTypes[2];
	int paramType;

	/* Set CUMULVS Options */

	stv_setopt( stvOptNull, stvOptVerbose, vflag ? STV_DEBUG_ALL : 0 );

	/* Initialize CUMULVS steering and visualization */

	stv_init( CPLANEGRP, STVTAG, NTASKS, myinst );

	/* Define the steering parameters */

	paramType = stvFloat;

	paramName = "Slope";
	paramPtr = (void *) &SLOPE;

	paramId[0] = stv_paramDefine( paramName, paramPtr, paramType,
		stvVisOnly );

	paramName = "Intercept";
	paramPtr = (void *) &INTERCEPT;

	paramId[1] = stv_paramDefine( paramName, paramPtr, paramType,
		stvVisOnly );

	paramName = "Rise";
	paramPtr = (void *) &RISE;

	paramId[2] = stv_paramDefine( paramName, paramPtr, paramType,
		stvVisOnly );

	paramName = "Exp";

	paramArray[0] = (void *) &BASE;
	paramNames[0] = "Base";
	paramTypes[0] = stvFloat;

	paramArray[1] = (void *) &FACTOR;
	paramNames[1] = "Factor";
	paramTypes[1] = stvFloat;

	paramId[3] = stv_vparamDefine( paramName,
		paramArray, paramNames, paramTypes, 2,
		stvVisOnly );

	paramType = stvInt;

	paramName = "Reset";
	paramPtr = (void *) &RESET;

	paramId[4] = stv_paramDefine( paramName, paramPtr, paramType,
		stvVisOnly );

	/* Declare a decomposition of density */
	/* -- it is block, cyclic, explicit */

	axisType[0] = stvBlock;
	axisInfo1[0] = stvDefault;

	axisType[1] = stvCyclic;
	axisInfo1[1] = CYCLESIZE;

	axisType[2] = stvExplicit;
	axisInfo1[2] = begink;
	axisInfo2[2] = endk;

	*decompId = stv_decompDefine( NDIM, axisType,
			axisInfo1, axisInfo2, glb, gub, PRANK, PSHAPE );

	/* declare the density data field with the above decomposition */

	declare[0] = IALLOC;
	declare[1] = JALLOC;
	declare[2] = KALLOC;

	offsets[0] = OFFSETI;
	offsets[1] = OFFSETJ;
	offsets[2] = OFFSETK;

	*fieldId = stv_fieldDefine( density, "density", *decompId,
			offsets, declare, stvFloat, paddr, stvVisOnly );

	stv_setopt( *fieldId, stvOptFieldTimeScale,
			stv_time_of( 100, stvTimeMicroSec ) );

	/* Verify Local Field Decomposition */

	stv_dumpLocalFieldBounds( *fieldId, stvVerbose );

	stv_dumpLocalFieldBounds( *fieldId, stvTerse );
}


void
bail()
{
	printf( "Exiting PVM...\n" );

	pvm_exit();

	printf( "Exiting.\n" );

	exit( -1 );
}

