
/* $Id: stv_steer.c,v 3.16 2000/07/18 21:35:58 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* System Header Files */

#include <stdio.h>
#include <math.h>


/* STV Header Files */

#include "stv.h"
#include "stvmac.h"
#include "stv_glob.h"


/*** Steering Routines ***/


/* Steering View Field Group Initialization */

int
stv_viewer_steering_init( V, freq, no_synch, restart )
STV_VIEWER V;
int freq;
int no_synch;
int *restart;
{
	static int first = 1;

	STV_VIEW_PARAM VP;

	int is_static = 0;
	int max_tids;
	int i, j;
	int flag;
	int cc;

	/* Add Steering Message Handlers */

	if ( first )
	{
		stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
				stv_protoSteerRelease, stvSteerRelease,
				stvAny, stvAny, STV_VIEWER_CTX );

		/* only the first time through! :-) */
		first = 0;
	}

	/* Clear Out Any Existing Steering View Field Group */

	if ( STV_STEER_VFG != NULL )
		stv_free_view_field_group( &STV_STEER_VFG );

	/* Create New Steering View Field Group Structure */

	STV_STEER_VFG = stv_create_view_field_group();

	STV_NO_MEM_return( STV_STEER_VFG, "stv_viewer_steering_init" );

	STV_STEER_VFG->fgid = STV_NEXT_FGID++;

	STV_STEER_VFG->nvfields = 0;

	/* Count Up Max View Param Tasks */

	VP = V->vparams;

	max_tids = 0;

	while ( VP != NULL )
	{
		max_tids += VP->ntids;

		VP = VP->next;
	}

	/* Verify Steering Possible */

	if ( max_tids > 0 )
	{
		/* Initialize Tasks List */

		STV_STEER_VFG->tid_list = (int *) MALLOC( (STV_SIZE_T) max_tids
			* sizeof( int ) );
		STV_MEMCHECK_return( STV_STEER_VFG->tid_list,
			"Steering View Field Group Task List Ptr" );

		STV_STEER_VFG->ntids = 0;

		/* Fill In Tasks */

		VP = V->vparams;

		while ( VP != NULL )
		{
			/* Get View Param Tasks */

			for ( i=0 ; i < VP->ntids ; i++ )
			{
				flag = 0;

				for ( j=0 ; j < STV_STEER_VFG->ntids && !flag ; j++ )
				{
					if ( STV_STEER_VFG->tid_list[j] == VP->tids[i] )
						flag++;
				}

				if ( !flag )
				{
					STV_STEER_VFG->tid_list
						[ (STV_STEER_VFG->ntids)++ ] = VP->tids[i];
				}
			}

			VP = VP->next;
		}

		STV_STEER_VFG->bail = stvFalse;

		/* Set Up Dummy Field Request */

		stv_init_region( STV_STEER_VFG->visregion );

		stv_init_cell( STV_STEER_VFG->viscell );

		STV_STEER_VFG->freq = freq;

		STV_STEER_VFG->eff_freq = stv_calculate_eff_freq(
			STV_STEER_VFG->freq, stv_time_of( 1, stvTimeNone ) );

		STV_STEER_VFG->flag = stvFieldRqstDefault;

		cc = stv_viewer_FieldRqst( STV_STEER_VFG, V,
			no_synch, is_static, restart );
	}

	else
		cc = stvStatusBadParam;

	return( cc );
}


/* Viewer Steering Request Routine */

int
stv_viewer_steering_request( VP, V, block )
STV_VIEW_PARAM VP;
STV_VIEWER V;
int block;
{
	char *group;

	int *tidlist;

	int ntids;
	int inst;
	int rbuf;
	int cc;

	/* Check for Null Args */

	if ( VP == NULL || V == NULL )
	{
		STV_ERROR( "stv_viewer_steering_request", stvStatusBadParam,
			"Null Argument Values" );
		
		return( stvStatusBadParam );
	}

	/* Verify Steering Initialized */

	if ( STV_STEER_VFG == NULL )
	{
		STV_ERROR( "stv_viewer_steering_request", stvStatusBadParam,
			"Viewer Steering Not Initialized" );

		return( stvStatusBadParam );
	}

	/* Check for Token - Already Snagged It, Bozo */

	if ( VP->token == stvSteerToken )
		return( stvStatusOk );

	/* Keep Trying Until Token Grabbed (Unless Not Blocking) */

	group = stv_steering_group( VP );

	do
	{
		/* Check for Steering Token Release */

		if ( VP->token == stvSteerRqstd )
		{
			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_VIEWER | STV_DEBUG_MSG ) )
			{
				printf( "%s for Steering Release Message.\n",
						( block == stvMsgBlock )
							? "Blocking" : "Probing" );
				fflush( stdout );
			}

			if ( (cc = stv_handleMessages( stvSteerRelease,
					stvAny, V->msgtag * 2, STV_VIEWER_CTX,
					block )) < 0 )
			{
				STV_ERROR( "stv_viewer_steering_request", cc,
						"Receiving Steering Release Message." );
				
				return( cc );
			}

			if ( cc > 0 )
			{
				VP->token = stvSteerNot;

				stv_leave_group( group, -1, -1 );
			}
		}

		/* Attempt to Grab Steering Token */

		if ( VP->token == stvSteerNot )
		{
			cc = stv_join_group( group, -1, &inst );

			if ( !cc )
			{
				stv_mp_err( "Joining Steering Token Group" );

				FREE( group );

				return( stvStatusBadParam );
			}

			/* Got It */

			if ( !inst )
			{
				VP->token = stvSteerToken;

				FREE( group );

				return( stvStatusOk );
			}

			/* Set Up Notifies & Hang Out */

			else
			{
				cc = stv_get_group_tids( group, tidlist, &ntids );

				if ( !cc )
				{
					stv_mp_err( "Obtaining Steering Token Group Tids" );

					FREE( group );

					return( stvStatusBadParam );
				}

				cc = stv_notify_event( stvTaskExit,
						V->msgtag * 2, STV_VIEWER_CTX,
						ntids, tidlist );

				if ( tidlist )
					FREE( tidlist );

				if ( !cc )
				{
					stv_mp_err( "Steering Token Group Notify" );

					FREE( group );

					return( stvStatusBadParam );
				}

				VP->token = stvSteerRqstd;

				if ( block != stvMsgBlock )
				{
					FREE( group );

					return( stvStatusBadParam );
				}
			}
		}
	}
	while ( block );

	FREE( group );

	return( stvStatusBadParam );
}


/* Steering Group Name Construction Routine */

char *stv_steering_group( VP )
STV_VIEW_PARAM VP;
{
	char groupname[1024];
	char tmp[255];

	char *valstr;

	int i;

	sprintf( groupname, "stv_steer_%s", VP->param->name );

	for ( i=0 ; i < VP->param->num ; i++ )
	{
		if ( VP->param->types[i] & stvIndex )
		{
			valstr = stv_str_of_typed_data( VP->param->vars[i],
				VP->param->types[i] & stvTypeMask );

			sprintf( tmp, "_%s", valstr );

			FREE( valstr );

			strcat( groupname, tmp );
		}
	}

	return( stv_copy_str( groupname ) );
}


/* Viewer Scalar Steering Parameter Update Routine */

int
stv_viewer_steer_parameter( VP, value, type )
STV_VIEW_PARAM VP;
STV_VALUE value;
int type;
{
	int cc;

	cc = stv_viewer_steer_vparameter( VP, value, type, 0 );

	if ( cc != stvStatusOk )
	{
		STV_ERROR( "stv_viewer_steer_parameter", stvStatusBadParam,
			"stv_viewer_steer_vparameter() failed." );

		return( stvStatusBadParam );
	}

	return( stvStatusOk );
}


/* Viewer Vector Steering Parameter Update Routine */

int
stv_viewer_steer_vparameter( VP, value, type, index )
STV_VIEW_PARAM VP;
STV_VALUE value;
int type;
int index;
{
	STV_PARAMPTR P;

	int cc;

	if ( VP == NULL || value == NULL )
	{
		STV_ERROR( "stv_viewer_steer_vparameter", stvStatusBadParam,
			"Null Argument Values" );
		
		return( stvStatusBadParam );
	}

	/* Verify Steering Initialized */

	if ( STV_STEER_VFG == NULL )
	{
		STV_ERROR( "stv_viewer_steer_vparameter", stvStatusBadParam,
			"Viewer Steering Not Initialized" );

		return( stvStatusBadParam );
	}

	/* Check Parameter Vector Index */

	P = VP->param;

	if ( index < 0 || index >= P->num )
	{
		STV_ERROR( "stv_viewer_steer_vparameter", stvStatusBadParam,
			"Parameter Index Out of Range" );

		return( stvStatusBadParam );
	}

	/* Verify Token Possession */

	if ( VP->token == stvSteerNot && !(P->types[ index ] & stvIndex) )
	{
		STV_ERROR( "stv_viewer_steer_vparameter", stvStatusBadParam,
			"Steering Token Not Acquired" );

		return( stvStatusBadParam );
	}

	/* Finally Copy the Bastard */

	cc = stv_copy_converted_data( value, type,
		P->vars[ index ], P->types[ index ] & stvTypeMask );
	
	if ( cc != stvStatusOk )
	{
		STV_ERROR( "stv_viewer_steer_vparameter", stvStatusBadParam,
			"stv_copy_converted_data() failed." );

		return( stvStatusBadParam );
	}

	VP->changed = stvTrue;

	return( stvStatusOk );
}


/* Viewer Steering Update Routine */

int
stv_viewer_send_NewParams( V )
STV_VIEWER V;
{
	STV_VIEW_PARAM VP;

	int sbuf;
	int tmp;
	int num;
	int cc;
	int i;

	/* Check for Null Args */

	if ( V == NULL )
	{
		STV_ERROR( "stv_viewer_send_NewParams", stvStatusBadParam,
			"Null Viewer Argument" );
		
		return( stvStatusBadParam );
	}

	/* Verify Steering Initialized */

	if ( STV_STEER_VFG == NULL )
	{
		STV_ERROR( "stv_viewer_send_NewParams", stvStatusBadParam,
			"Viewer Steering Not Initialized" );

		return( stvStatusBadParam );
	}

	if ( STV_OPT_VERBOSE & STV_DEBUG_STEER )
	{
		printf( "Sending Parameter Update:  Region Time = " );
		stv_dump_time_str( stdout,
			STV_STEER_VFG->region_time, stvTrue );
		fflush( stdout );
	}

	stv_init_msg_buf( stvDataDefault, &sbuf );

	tmp = stvMsgOp( stvNewParams );
	stv_pack_int( sbuf, &tmp, 1, 1 );

	stv_pack_int( sbuf, &(STV_STEER_VFG->fgid), 1, 1 );

	stv_pack_int( sbuf, &(STV_STEER_VFG->region_time), 1, 1 );

	/* Pack Up Number of Modified Params */

	VP = V->vparams;

	num = 0;

	while ( VP != NULL )
	{
		if ( VP->changed )
		{
			if ( VP->token != stvSteerToken )
			{
				sprintf( STVERRTXT, "Warning: No Steering Token for %s",
					VP->param->name );

				STV_ERROR( "stv_viewer_send_NewParams",
					stvStatusBadParam, STVERRTXT );
			}

			else
				num++;
		}

		VP = VP->next;
	}

	stv_pack_int( sbuf, &(num), 1, 1 );

	/* Pack Up Yer Troubles */

	VP = V->vparams;

	while ( VP != NULL )
	{
		if ( VP->changed && VP->token == stvSteerToken )
		{
			if ( STV_OPT_VERBOSE & STV_DEBUG_STEER )
			{
				printf( "\t%s\n", VP->param->name );
				fflush( stdout );
			}

			stv_pack_int( sbuf, &(VP->param->paramID), 1, 1 );

			for ( i=0 ; i < VP->param->num ; i++ )
			{
				if ( ( STV_OPT_VERBOSE & STV_DEBUG_STEER )
						&& VP->param->num != 1 )
				{
					printf( "\t\t%s\n", VP->param->pnames[i] );
					fflush( stdout );
				}

				if ( (cc = stv_pack_by_type( sbuf, VP->param->vars[i],
					1, 1, VP->param->types[i] & stvTypeMask ))
					!= stvStatusOk )
				{
					STV_ERROR( "stv_viewer_send_NewParams", cc,
						"Typed Pack Failed." );

					return( cc );
				}
			}
		}

		VP->changed = stvFalse;

		VP = VP->next;
	}

	stv_mcast_msg( sbuf, STV_STEER_VFG->tid_list, STV_STEER_VFG->ntids,
		V->msgtag, STV_APP_CTX );

	stv_free_msg_buf( sbuf );

	return( stvStatusOk );
}


/* Viewer Steering Release Routine */

int
stv_viewer_steering_release( VP, V )
STV_VIEW_PARAM VP;
STV_VIEWER V;
{
	char *group;

	int *tidlist;

	int ntids;
	int sbuf;
	int tmp;
	int cc;

	/* Check for Null Args */

	if ( VP == NULL || V == NULL )
	{
		STV_ERROR( "stv_viewer_steering_release", stvStatusBadParam,
			"Null Argument Values" );
		
		return( stvStatusBadParam );
	}

	/* Verify Steering Initialized */

	if ( STV_STEER_VFG == NULL )
	{
		STV_ERROR( "stv_viewer_steering_release", stvStatusBadParam,
			"Viewer Steering Not Initialized" );

		return( stvStatusBadParam );
	}

	/* Verify Token Possession */

	if ( VP->token == stvSteerNot )
	{
		STV_ERROR( "stv_viewer_steering_release", stvStatusBadParam,
			"Warning: No Steering Token to Release" );

		return( stvStatusBadParam );
	}

	/* Release Token */

	group = stv_steering_group( VP );

	if ( VP->token == stvSteerToken )
	{
		stv_init_msg_buf( stvDataDefault, &sbuf );

		tmp = stvMsgOp( stvSteerRelease );
		stv_pack_int( sbuf, &tmp, 1, 1 );

		/* dummy fgid, for message header consistency */
		tmp = -1;
		stv_pack_int( sbuf, &tmp, 1, 1 );

		stv_broadcast( sbuf, group, V->msgtag * 2, STV_VIEWER_CTX );
	}

	stv_leave_group( group, -1, -1 );

	FREE( group );

	VP->token = stvSteerNot;

	return( stvStatusOk );
}


/* Receive and Process One Steering Release Message */

int
stv_protoSteerRelease( msg )
STV_BUFPTR msg;
{
	if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_MSG ) )
	{
		printf( "Entry stv_protoSteerRelease().\n" );
		fflush( stdout );
	}

	if ( msg == (STV_BUFPTR) NULL )
		return( stvStatusOk );
	
	/* Don't Do Anything Specific, Just Know That Someone Released */
	/* Any pending steering requests will now re-try... */

	return( stvStatusOk );
}

